import subprocess  # nosec
from pathlib import Path
from typing import Optional, Union

from pathspec import PathSpec

from ._minify import minify as minify_code
from .hash import fnv1a
from .typing import PathType


def discover_files_dirs(
    remote_dir: str,
    local_file_or_folder: Path,
    ignore: Optional[list] = None,
):
    src_objects = []
    if local_file_or_folder.is_dir():
        if ignore is None:
            ignore = []
        ignore_spec = PathSpec.from_lines("gitwildmatch", ignore)
        for src_object in local_file_or_folder.rglob("*"):
            if ignore_spec.match_file(str(src_object)):
                continue
            src_objects.append(src_object)
        # Sort so that folder creation comes before file sending.
        src_objects.sort()

        src_files, src_dirs = [], []
        for src_object in src_objects:
            if src_object.is_dir():
                src_dirs.append(src_object)
            else:
                src_files.append(src_object)
        dst_files = [remote_dir / src.relative_to(local_file_or_folder) for src in src_files]
    else:
        src_files = [local_file_or_folder]
        src_dirs = []
        dst_files = [Path(remote_dir) / local_file_or_folder.name]

    return src_files, src_dirs, dst_files


def preprocess_keep(
    keep: Union[None, list, str, bool],
    dst: str,
) -> list:
    if keep is None:
        keep = ["boot.py", "webrepl_cfg.py", "lib"] if dst == "/" else []
    elif isinstance(keep, str):
        keep = [keep]
    elif isinstance(keep, (list, tuple)):
        pass
    elif isinstance(keep, bool):
        keep = []
    else:
        raise TypeError
    keep = [(dst / Path(x)).as_posix() for x in keep]
    return keep


def preprocess_ignore(ignore: Union[None, str, list, tuple]) -> list:
    if ignore is None:
        ignore = ["*.pyc", "__pycache__", ".DS_Store", ".pytest_cache"]
    elif isinstance(ignore, str):
        ignore = [ignore]
    elif isinstance(ignore, (list, tuple)):
        ignore = list(ignore)
    else:
        raise TypeError
    return ignore


def preprocess_src_file(
    tmp_dir: PathType,
    src_file: PathType,
    minify: bool,
    mpy_cross_binary: Union[str, Path, None],
) -> Path:
    tmp_dir = Path(tmp_dir)
    src_file = Path(src_file)

    transformed = tmp_dir / src_file.relative_to(tmp_dir.anchor) if src_file.is_absolute() else tmp_dir / src_file
    transformed.parent.mkdir(parents=True, exist_ok=True)

    if src_file.suffix == ".py":
        if mpy_cross_binary:
            transformed = transformed.with_suffix(".mpy")
            subprocess.check_output([mpy_cross_binary, "-o", transformed, src_file])  # nosec
            return transformed
        elif minify:
            minified = minify_code(src_file.read_text())
            transformed.write_text(minified)
            return transformed

    return src_file


def preprocess_src_file_hash(*args, **kwargs):
    src_file = preprocess_src_file(*args, **kwargs)
    src_hash = fnv1a(src_file)
    return src_file, src_hash


def generate_dst_dirs(dst, src, src_dirs) -> list:
    dst_dirs = [(dst / x.relative_to(src)).as_posix() for x in src_dirs]
    # Add all directories leading up to ``dst``.
    dst_prefix_tokens = dst.split("/")
    for i in range(2, len(dst_prefix_tokens) + (dst[-1] != "/")):
        dst_dirs.append("/".join(dst_prefix_tokens[:i]))
    dst_dirs.sort()
    return dst_dirs



from typing import Dict, List, Optional

from pydantic import BaseModel, Field
from serial.tools.list_ports import comports

from .exceptions import DeviceNotFoundError, InsufficientSpecifierError


def _normalize(val):
    """Normalize ``val`` for comparison."""
    if isinstance(val, str):
        return val.casefold()
    return val


def _dict_is_subset(subset: Dict, superset: Dict) -> bool:
    """Tests if ``subset`` dictionary is a subset of ``superset`` dictionary."""
    for subset_key, subset_value in subset.items():
        try:
            superset_value = superset[subset_key]
        except KeyError:
            return False

        if _normalize(subset_value) != _normalize(superset_value):
            return False
    return True


class UsbSpecifier(BaseModel):
    """Usb port metadata."""

    vid: Optional[int] = None
    pid: Optional[int] = None
    serial_number: Optional[str] = None
    manufacturer: Optional[str] = None
    product: Optional[str] = None
    location: Optional[str] = None

    device: Optional[str] = Field(None, exclude=True)

    def __repr__(self):
        return f'{self.__class__.__name__}({", ".join(f"{k}={v!r}" for k, v in self.dict().items() if v is not None)})'

    def to_port(self) -> str:
        if self.device:
            return self.device

        spec = self.dict(exclude_none=True)
        possible_matches = []

        for port_info in list_devices():
            if _dict_is_subset(spec, vars(port_info)):
                possible_matches.append(port_info)
        if not possible_matches:
            raise DeviceNotFoundError
        elif len(possible_matches) > 1:
            message = "Multiple potential devices found:\n" + "\n".join(f"    {vars(x)}" for x in possible_matches)
            raise InsufficientSpecifierError(message)

        return possible_matches[0].device

    def populated(self):
        # some ports, like wlan and bluetooth on macos,
        # don't populate any meaningful fields.
        return bool(self.dict(exclude_none=True))


def list_devices() -> List[UsbSpecifier]:
    """Lists available device ports.

    Returns
    -------
    List[UsbSpecifier]
        Available devices identifiers.
    """
    devices = [
        UsbSpecifier(
            vid=port.vid,
            pid=port.pid,
            serial_number=port.serial_number,
            manufacturer=port.manufacturer,
            product=port.product,
            location=port.location,
            device=port.device,
        )
        for port in comports()
    ]
    return [x for x in devices if x.populated()]



from pathlib import Path

from typer import Argument, Option

from belay import Device
from belay.cli.common import help_password, help_port, remove_stacktrace


def run(
    port: str = Argument(..., help=help_port),
    file: Path = Argument(..., help="File to run on-device."),
    password: str = Option("", help=help_password),
):
    """Run file on-device.

    If the first argument, ``port``, is resolvable to an executable,
    the remainder of the command will be interpreted as a shell command
    that will be executed in a pseudo-micropython-virtual-environment.
    As of right now, this just sets ``MICROPYPATH`` to all of the dependency
    groups' folders. E.g::

            belay run micropython -m unittest

    """
    content = file.read_text()
    with Device(port, password=password) as device, remove_stacktrace():
        device(content)



import shutil
from functools import partial
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Optional

from rich.progress import Progress
from typer import Argument, Option

from belay import Device
from belay.cli.common import help_password, help_port, remove_stacktrace
from belay.cli.sync import sync_device as _sync_device
from belay.project import find_project_folder, load_groups, load_pyproject


def install(
    port: str = Argument(..., help=help_port),
    password: str = Option("", help=help_password),
    mpy_cross_binary: Optional[Path] = Option(None, help="Compile py files with this executable."),
    run: Optional[Path] = Option(None, help="Run script on-device after installing."),
    main: Optional[Path] = Option(None, help="Sync script to /main.py after installing."),
    with_groups: List[str] = Option(None, "--with", help="Include specified optional dependency group."),
    follow: bool = Option(False, "--follow", "-f", help="Follow the stdout after upload."),
):
    """Sync dependencies and project itself to device."""
    if run and run.suffix != ".py":
        raise ValueError("Run script MUST be a python file.")
    if main and main.suffix != ".py":
        raise ValueError("Main script MUST be a python file.")

    config = load_pyproject()
    project_folder = find_project_folder()
    project_package = config.name
    groups = load_groups()

    with Device(port, password=password) as device:
        sync_device = partial(
            _sync_device,
            device,
            mpy_cross_binary=mpy_cross_binary,
        )

        with TemporaryDirectory() as tmp_dir, Progress() as progress:
            tmp_dir = Path(tmp_dir)

            # Add all tasks to progress bar
            tasks = {}

            def create_task(key, task_description):
                task_id = progress.add_task(task_description)

                def progress_update(description=None, **kwargs):
                    if description:
                        description = task_description + description
                    progress.update(task_id, description=description, **kwargs)

                tasks[key] = progress_update

            create_task("dependencies", "Dependencies: ")
            if project_package:
                create_task("project_package", f"{project_package}: ")
            if main:
                create_task("main", "main: ")

            # Aggregate dependencies to an intermediate temporary directory.
            for group in groups:
                if group.optional and group.name not in with_groups:
                    continue
                group.copy_to(tmp_dir)

            sync_device(
                tmp_dir,
                dst="/lib",
                progress_update=tasks["dependencies"],
            )

            if project_package:
                sync_device(
                    folder=project_folder / project_package,
                    dst=f"/{project_package}",
                    progress_update=tasks["project_package"],
                    ignore=config.ignore,
                )

            if main:
                # Copy provided main to temporary directory in case it's not named main.py
                main_tmp = tmp_dir / "main.py"
                shutil.copy(main, main_tmp)
                sync_device(main_tmp, progress_update=tasks["main"])

        if run:
            content = run.read_text()
            with remove_stacktrace():
                device(content)
            return

        # Reset device so ``main.py`` has a chance to execute.
        device.soft_reset()
        if follow:
            device.terminal()



import math
from dataclasses import dataclass
from threading import Lock
from typing import Callable, Optional, Tuple


@dataclass
class Implementation:
    """Implementation dataclass detailing the device.

    Parameters
    ----------
    name: str
        Type of python running on device.
        One of ``{"micropython", "circuitpython"}``.
    version: Tuple[int, int, int]
        ``(major, minor, patch)`` Semantic versioning of device's firmware.
    platform: str
        Board identifier. May not be consistent from MicroPython to CircuitPython.
        e.g. The Pi Pico is "rp2" in MicroPython, but "RP2040"  in CircuitPython.
    emitters: tuple[str]
        Tuple of available emitters on-device ``{"native", "viper"}``.
    """

    name: str
    version: Tuple[int, int, int] = (0, 0, 0)
    platform: str = ""
    emitters: Tuple[str] = ()


_method_metadata_counter_lock = Lock()
_method_metadata_counter = 0


@dataclass
class MethodMetadata:
    """Metadata for executer-decorated Device methods."""

    executer: Callable
    kwargs: dict
    autoinit: bool = False  # Only applies to ``SetupExecuter``.
    implementation: Optional[str] = None

    id: int = -1  # monotonically increasing global identifier.

    def __post_init__(self):
        global _method_metadata_counter
        with _method_metadata_counter_lock:
            self.id = _method_metadata_counter
            _method_metadata_counter += 1


def sort_executers(executers):
    """Sorts executers by monotonically increasing ``__belay__.id``.

    This ensures that, when necessary, executers are called in the order they are defined.
    """

    def get_key(x):
        try:
            return x.__wrapped__.__belay__.id
        except AttributeError:
            return math.inf

    return sorted(executers, key=get_key)



class BelayException(Exception):  # noqa: N818
    """Root Belay exception class."""


class AuthenticationError(BelayException):
    """Invalid password or similar."""


class FeatureUnavailableError(BelayException):
    """Feature unavailable for your board's implementation."""


class SpecialFunctionNameError(BelayException):
    """Attempted to use a reserved Belay function name.

    The following name rules are reserved:

        * Names that start and end with double underscore, ``__``.

        * Names that start with ``_belay`` or ``__belay``
    """


class MaxHistoryLengthError(BelayException):
    """Too many commands were given."""


class DeviceNotFoundError(BelayException):
    """Unable to find specified device."""


class InsufficientSpecifierError(BelayException):
    """Specifier wasn't unique enough to determine a single device."""


class ConnectionFailedError(BelayException):
    """Unable to connect to specified device."""


class ConnectionLost(ConnectionFailedError):  # noqa: N818
    """Lost connection to device."""


class InternalError(BelayException):
    """Internal to Belay logic error."""


class NotBelayResponseError(BelayException):
    """Parsed response wasn't for Belay."""


class NoMatchingExecuterError(BelayException):
    """No valid executer found for the given board Implementation."""



import importlib.resources as pkg_resources
import re
import shutil
from pathlib import Path

from packaging.utils import canonicalize_name
from typer import Argument, Option


def new(project_name: str = Argument(..., help="Project Name.")):
    """Create a new micropython project structure."""
    package_name = canonicalize_name(project_name)
    dst_dir = Path() / project_name
    template_dir = pkg_resources.files("belay") / "cli" / "new_template"

    shutil.copytree(str(template_dir), str(dst_dir))

    # Find/Replace Engine
    replacements: dict[str, str] = {
        "packagename": package_name,
    }

    paths = list(dst_dir.rglob("*"))

    def replace(string):
        """Replace whole words only."""

        def _replace(match):
            return replacements[match.group(0)]

        return re.sub("|".join(r"\b%s\b" % re.escape(s) for s in replacements), _replace, string)

    for path in paths:
        if path.is_dir():
            continue

        contents = path.read_text()
        contents = replace(contents)
        path.write_text(contents)
        if path.stem in replacements:
            dst = path.with_name(replacements[path.stem] + path.suffix)
            path.replace(dst)

    # Move the app folder
    (dst_dir / "packagename").replace(dst_dir / replacements["packagename"])



from contextlib import nullcontext
from functools import partial
from pathlib import Path
from typing import List, Optional

from rich.progress import Progress
from typer import Argument, Option

from belay import Device
from belay.cli.common import help_password, help_port


def sync_device(device, folder, progress_update, **kwargs):
    device.sync(folder, progress_update=progress_update, **kwargs)
    progress_update(description="Complete.")


def sync(
    port: str = Argument(..., help=help_port),
    folder: Path = Argument(..., help="Path of local file or folder to sync."),
    dst: str = Option("/", help="Destination directory to unpack folder contents to."),
    password: str = Option("", help=help_password),
    keep: Optional[List[str]] = Option(None, help="Files to keep."),
    ignore: Optional[List[str]] = Option(None, help="Files to ignore."),
    mpy_cross_binary: Optional[Path] = Option(None, help="Compile py files with this executable."),
):
    """Synchronize a folder to device."""
    # Typer issues: https://github.com/tiangolo/typer/issues/410
    keep = keep if keep else None
    ignore = ignore if ignore else None

    with Device(port, password=password) as device, Progress() as progress:
        task_id = progress.add_task("")

        def progress_update(description=None, **kwargs):
            return progress.update(task_id, description=description, **kwargs)

        sync_device(
            device,
            folder,
            progress_update,
            dst=dst,
            keep=keep,
            ignore=ignore,
            mpy_cross_binary=mpy_cross_binary,
        )



from typing import List

from rich.console import Console
from typer import Argument

from belay.cli.clean import clean
from belay.project import load_groups


def update(packages: List[str] = Argument(None, help="Specific package(s) to update.")):
    """Download new versions of dependencies."""
    console = Console()
    groups = load_groups()
    packages = packages if packages else None

    for group in groups:
        group_packages = None if packages is None else [x for x in packages if x in group.config.dependencies]

        group.download(
            packages=group_packages,
            console=console,
        )

    clean()



import ast
import shutil
import tempfile
from contextlib import nullcontext
from pathlib import Path
from typing import List, Optional

from rich.console import Console

from belay.packagemanager.downloaders import download_uri
from belay.packagemanager.models import (
    DependencySourceConfig,
    GroupConfig,
    walk_dependencies,
)
from belay.packagemanager.sync import sync
from belay.typing import PathType


class Group:
    """Represents a group defined in ``pyproject.toml``."""

    def __init__(self, name: str, **kwargs):
        from belay.project import find_dependencies_folder

        self.name = name
        self.config = GroupConfig(**kwargs)
        self.folder = find_dependencies_folder() / self.name

    def __eq__(self, other):
        if not isinstance(other, Group):
            return False
        return self.config.__dict__ == other.config.__dict__

    def __repr__(self):
        kws = [f"{key}={value!r}" for key, value in self.config.__dict__.items()]
        return f"{type(self).__name__}({', '.join(kws)})"

    @property
    def optional(self) -> bool:
        return self.config.optional

    @property
    def dependencies(self):
        return self.config.dependencies

    def clean(self):
        """Delete any dependency module not specified in ``self.config.dependencies``."""
        dependencies = set(self.dependencies)

        if not self.folder.exists():
            return

        for existing_dep in self.folder.glob("*"):
            if existing_dep.name in dependencies:
                continue

            if existing_dep.is_dir():
                shutil.rmtree(existing_dep)
            else:
                existing_dep.unlink()

    def copy_to(self, dst: PathType) -> None:
        """Copy Dependencies folder to destination directory.

        Used to stage files for sync/installation.
        """
        dst = Path(dst)

        if self.folder.exists():
            # Bulk copy over the group contents
            shutil.copytree(self.folder, dst, dirs_exist_ok=True)

        # Copy over any (& overwrite) any dependencies in ``develop`` mode.
        for package_name, dependency in _walk_develop_dependencies(self.dependencies):
            dst_package_folder = dst / package_name
            dst_package_folder.mkdir(parents=True, exist_ok=True)
            _download_and_verify_dependency(dst_package_folder, dependency)

    def _download_package(self, package_name) -> bool:
        """Download a single package.

        Returns
        -------
        changed: bool
            If existing files have changed after download.
        """
        local_folder = self.folder / package_name
        local_folder.mkdir(exist_ok=True, parents=True)

        dependencies = self.dependencies[package_name]

        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            for dependency in dependencies:
                _download_and_verify_dependency(tmp_dir, dependency)
            changed = sync(tmp_dir, local_folder)

        return changed

    def download(
        self,
        packages: Optional[List[str]] = None,
        console: Optional[Console] = None,
    ) -> None:
        """Download dependencies.

        Parameters
        ----------
        packages: Optional[List[str]]
            Only download these package.
        console: Optional[Console]
            Print progress out to console.
        """
        if packages is None:
            # Update all packages
            packages = list(self.dependencies.keys())

        if not packages:
            return

        cm = console.status("[bold green]Updating Dependencies") if console else nullcontext()

        def log(*args, **kwargs):
            if console:
                console.print(*args, **kwargs)

        with cm:
            for package_name in packages:
                log(f"  • {package_name}: Updating...", end=" ")
                changed = self._download_package(package_name)
                if changed:
                    log(f"  • [bold green]{package_name}: Updated.")
                else:
                    log(f"  • {package_name}: No changes detected.")


def _verify_files(path: PathType):
    """Sanity checks downloaded files.

    Performs the following checks:

    * ".py" files are valid python code.

    Parameters
    ----------
    path
        Either a single file or a folder.
    """
    path = Path(path)

    gen = path.rglob("*") if path.is_dir() else [path]

    for f in gen:
        if f.suffix == ".py":
            code = f.read_text()
            ast.parse(code)


def _walk_develop_dependencies(packages: dict):
    for package_name, dependency in walk_dependencies(packages):
        if dependency.develop:
            yield package_name, dependency


def _download_and_verify_dependency(download_folder: PathType, dependency: DependencySourceConfig):
    """Download and verify a dependency.

    Parameters
    ----------
    download_folder
        Destination directory for downloaded file(s).
    """
    out = download_uri(download_folder, dependency.uri)
    if dependency.rename_to_init and out.is_file() and out.suffix == ".py":
        out = out.rename(out.parent / "__init__.py")

    _verify_files(out)



from belay.pyboard import PyboardException

help_port = "Port (like /dev/ttyUSB0) or WebSocket (like ws://192.168.1.100) of device."
help_password = "Password for communication methods (like WebREPL) that require authentication."  # nosec  # noqa: S105


class remove_stacktrace:  # noqa: N801
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None and issubclass(exc_type, PyboardException):
            print(exc_value)
            return True  # suppress the full stack trace
        else:
            return False  # let other exceptions propagate normally



import re
import shutil
from pathlib import Path

import git
import requests

from .common import NonMatchingURI, downloaders


@downloaders
def github(dst: Path, uri: str):
    """Download a file or folder from github."""
    # Single File Website; e.g.:
    #     https://github.com/BrianPugh/belay/blob/main/belay/__init__.py
    match = re.search(r"github\.com/(.+?)/(.+?)/blob/(.+?)/(.*)", uri)
    if not match:
        # Folder; e.g.:
        #     https://github.com/BrianPugh/belay/tree/main/belay
        match = re.search(r"github\.com/(.+?)/(.+?)/tree/(.+?)/(.*)", uri)
    if not match:
        match = re.search(r"raw\.githubusercontent\.com/(.+?)/(.+?)/(.+?)/(.*)", uri)
    if not match:
        raise NonMatchingURI
    org, repo, ref, path = match.groups()

    githubusercontent_url = f"https://raw.githubusercontent.com/{org}/{repo}/{ref}/{path}"

    r = requests.get(githubusercontent_url, timeout=10.0)

    if r.status_code == 200:
        # Provided URI is a single file.
        dst /= Path(path).name
        dst.write_bytes(r.content)
    elif r.status_code == 404:
        # Probably a folder; use git.
        from belay.project import find_cache_folder

        repo_url = f"https://github.com/{org}/{repo}.git"
        repo_folder = find_cache_folder() / f"git-github-{org}-{repo}"
        repo_folder.mkdir(exist_ok=True, parents=True)

        # Check if we have already cloned
        if (repo_folder / ".git").is_dir():
            # Already been cloned
            repo = git.Repo(repo_folder)
            repo.remotes.origin.pull()
        else:
            repo = git.Repo.clone_from(repo_url, repo_folder)

        repo.git.clean("-xdf")
        repo.git.checkout(ref)

        shutil.copytree(repo_folder / path, dst, dirs_exist_ok=True)
    else:
        r.raise_for_status()

    return dst



from typing import Any, Dict, Optional, Sequence, Union

from prompt_toolkit import PromptSession
from prompt_toolkit.application import Application
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.styles import Style, merge_styles
from questionary import utils
from questionary.constants import (
    DEFAULT_QUESTION_PREFIX,
    DEFAULT_SELECTED_POINTER,
    DEFAULT_STYLE,
)
from questionary.prompts import common
from questionary.prompts.common import Choice, InquirerControl
from questionary.question import Question


def press_any_key_to_continue(
    message: str = "Press any key to continue...",
    style: Optional[Style] = None,
    **kwargs,
):
    """Wait until user presses any key to continue.

    Example:
        >>> import questionary
        >>> questionary.press_any_key_to_continue().ask()
         Press any key to continue...
        None

    Args:
        message: Question text.

        style: A custom color and style for the question parts. You can
               configure colors as well as font types for different elements.
    """
    merged_style = merge_styles([DEFAULT_STYLE, style])

    def get_prompt_tokens():
        tokens = []

        tokens.append(("class:question", f" {message} "))

        return to_formatted_text(tokens)

    def exit_with_result(event):
        event.app.exit(result=None)

    bindings = KeyBindings()

    @bindings.add(Keys.Any)
    def any_key(event):
        exit_with_result(event)

    return Question(PromptSession(get_prompt_tokens, key_bindings=bindings, style=merged_style, **kwargs).app)


def select_table(
    message: str,
    header: str,
    choices: Sequence[Union[str, Choice, Dict[str, Any]]],
    default: Optional[Union[str, Choice, Dict[str, Any]]] = None,
    qmark: str = DEFAULT_QUESTION_PREFIX,
    pointer: Optional[str] = DEFAULT_SELECTED_POINTER,
    style: Optional[Style] = None,
    use_indicator: bool = False,
    **kwargs: Any,
) -> Question:
    """A list of items to select **one** option from.

    Simplified to work better with a formatted table.

    Args:
        message: Question text
        header: Table header text
        choices: Items shown in the selection, this can contain :class:`Choice` or
                 or :class:`Separator` objects or simple items as strings. Passing
                 :class:`Choice` objects, allows you to configure the item more
                 (e.g. preselecting it or disabling it).
        default: A value corresponding to a selectable item in the choices,
                 to initially set the pointer position to.
        qmark: Question prefix displayed in front of the question.
               By default this is a ``?``.
        pointer: Pointer symbol in front of the currently highlighted element.
                 By default this is a ``»``.
                 Use ``None`` to disable it.
        style: A custom color and style for the question parts. You can
               configure colors as well as font types for different elements.
        use_indicator: Flag to enable the small indicator in front of the
                       list highlighting the current location of the selection
                       cursor.

    Returns
    -------
        :class:`Question`: Question instance, ready to be prompted (using ``.ask()``).
    """
    if choices is None or len(choices) == 0:
        raise ValueError("A list of choices needs to be provided.")

    merged_style = merge_styles([DEFAULT_STYLE, style])

    ic = InquirerControl(
        choices,
        default,
        pointer=pointer,
        use_indicator=use_indicator,
        initial_choice=default,
    )

    def get_prompt_tokens():
        # noinspection PyListCreation
        tokens = [
            ("class:qmark", qmark),
            ("class:question", f" {message} "),
            ("class:question", "\n   " + header),
        ]

        if ic.is_answered:
            tokens.append(("class:answer", "\n   " + ic.get_pointed_at().title))

        return tokens

    layout = common.create_inquirer_layout(ic, get_prompt_tokens, **kwargs)

    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def _(event):
        event.app.exit(exception=KeyboardInterrupt, style="class:aborting")

    @bindings.add(Keys.Down, eager=True)
    @bindings.add("j", eager=True)
    @bindings.add(Keys.ControlN, eager=True)
    def move_cursor_down(event):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add("k", eager=True)
    @bindings.add(Keys.ControlP, eager=True)
    def move_cursor_up(event):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event):
        ic.is_answered = True
        event.app.exit(result=ic.get_pointed_at().value)

    @bindings.add(Keys.Any)
    def other(event):
        """Disallow inserting other text."""

    return Question(
        Application(
            layout=layout,
            key_bindings=bindings,
            style=merged_style,
            **utils.used_kwargs(kwargs, Application.__init__),
        )
    )



import os, time, machine
from time import sleep
from micropython import const
from machine import Pin, PWM, Timer
try:
    from machine import I2C
except ImportError:
    pass
try:
    from machine import SPI
except ImportError:
    pass
try:
    from machine import ADC
except ImportError:
    pass



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2014-2021 Damien P. George
# Copyright (c) 2017 Paul Sokolovsky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
pyboard interface.

This module provides the Pyboard class, used to communicate with and
control a MicroPython device over a communication channel. Both real
boards and emulated devices (e.g. running in QEMU) are supported.
Various communication channels are supported, including a serial
connection, telnet-style network connection, external process
connection.

Example usage:

    import pyboard
    pyb = pyboard.Pyboard('/dev/ttyACM0')

Or:

    pyb = pyboard.Pyboard('192.168.1.1')

Then:

    pyb.enter_raw_repl()
    pyb.exec('import pyb')
    pyb.exec('pyb.LED(1).on()')
    pyb.exit_raw_repl()
"""

import ast
import atexit
import contextlib
import itertools
import os
import platform
import signal
import subprocess
import sys
import time
from pathlib import Path
from threading import Lock, Thread
from typing import Union

from pydantic import ValidationError

from .exceptions import BelayException, ConnectionFailedError, DeviceNotFoundError
from .usb_specifier import UsbSpecifier
from .webrepl import WebreplToSerial

try:
    stdout = sys.stdout.buffer
except AttributeError:
    # Python2 doesn't have buffer attr
    stdout = sys.stdout


def _kill_process(pid):
    try:
        if platform.system() == "Windows":
            subprocess.run(["taskkill", "/F", "/T", "/PID", str(pid)])
        else:
            os.killpg(os.getpgid(pid), signal.SIGTERM)
    except ProcessLookupError:
        pass


def _parse_bool(env_var, default_value=False):
    if env_var in os.environ:
        env_value = os.environ[env_var].lower()
        return env_value == "true" or env_value == "1"
    else:
        return default_value


def stdout_write_bytes(b):
    b = b.replace(b"\x04", b"")
    stdout.write(b)
    stdout.flush()


def _dummy_data_consumer(data):
    pass


class PyboardError(BelayException):
    """An issue communicating with the board or parsing board response."""


class PyboardException(BelayException):  # noqa: N818
    """Uncaught exception from user code on the device."""

    def __str__(self):
        return "\n\n" + self.args[0]


class TelnetToSerial:
    def __init__(self, ip, user, password, read_timeout=None):
        self.tn = None
        import telnetlib

        self.tn = telnetlib.Telnet(ip, timeout=15)
        self.read_timeout = read_timeout
        if b"Login as:" in self.tn.read_until(b"Login as:", timeout=read_timeout):
            self.tn.write(bytes(user, "ascii") + b"\r\n")

            if b"Password:" in self.tn.read_until(b"Password:", timeout=read_timeout):
                # needed because of internal implementation details of the telnet server
                time.sleep(0.2)
                self.tn.write(bytes(password, "ascii") + b"\r\n")

                if b"for more information." in self.tn.read_until(
                    b'Type "help()" for more information.', timeout=read_timeout
                ):
                    # login successful
                    from collections import deque

                    self.fifo = deque()
                    return

        raise ConnectionFailedError

    def __del__(self):
        self.close()

    def close(self):
        if self.tn:
            self.tn.close()

    def read(self, size=1):
        while len(self.fifo) < size:
            timeout_count = 0
            data = self.tn.read_eager()
            if len(data):
                self.fifo.extend(data)
                timeout_count = 0
            else:
                time.sleep(0.25)
                if self.read_timeout is not None and timeout_count > 4 * self.read_timeout:
                    break
                timeout_count += 1

        data = b""
        while len(data) < size and len(self.fifo) > 0:
            data += bytes([self.fifo.popleft()])
        return data

    def write(self, data):
        self.tn.write(data)
        return len(data)

    @property
    def in_waiting(self):
        n_waiting = len(self.fifo)
        if not n_waiting:
            data = self.tn.read_eager()
            self.fifo.extend(data)
            return len(data)
        else:
            return n_waiting


class ProcessToSerial:
    """Execute a process and emulate serial connection using its stdin/stdout."""

    def __init__(self, cmd):
        import subprocess

        self.subp = subprocess.Popen(
            cmd,
            bufsize=0,
            shell=True,
            start_new_session=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        time.sleep(0.5)

        self.buf = bytearray()
        self.lock = Lock()

        def process_output():
            assert self.subp.stdout is not None  # noqa: S101
            while True:
                out = self.subp.stdout.read(1)
                if out:
                    with self.lock:
                        self.buf.extend(out)
                elif self.subp.poll() is not None:
                    break

        thread = Thread(target=process_output)
        thread.daemon = True
        thread.start()

        atexit.register(self.close)

        t_deadline = time.time() + 1
        while b">>>" not in self.buf:
            time.sleep(0.0001)
            if _parse_bool("BELAY_DEBUG_PROCESS_BUFFER") and time.time() > t_deadline:
                t_deadline = time.time() + 1
                print(self.buf)

    def close(self):
        _kill_process(self.subp.pid)
        atexit.unregister(self.close)

    def read(self, size=1):
        while len(self.buf) < size:
            # yield to the reading threads
            time.sleep(0.0001)

        with self.lock:
            data = self.buf[:size]
            self.buf[:] = self.buf[size:]

        return data

    def write(self, data):
        self.subp.stdin.write(data)
        return len(data)

    @property
    def in_waiting(self):
        return len(self.buf)


class ProcessPtyToTerminal:
    """Creates a PTY process and prints slave PTY as first line of its output.

    Emulates serial connection using this PTY.
    """

    def __init__(self, cmd):
        import re
        import subprocess

        import serial

        self.subp = subprocess.Popen(
            cmd.split(),
            bufsize=0,
            shell=False,
            start_new_session=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        pty_line = self.subp.stderr.readline().decode("utf-8")
        m = re.search(r"/dev/pts/[0-9]+", pty_line)
        if not m:
            print("Error: unable to find PTY device in startup line:", pty_line)
            self.close()
            sys.exit(1)
        pty = m.group()
        # rtscts, dsrdtr params are to workaround pyserial bug:
        # http://stackoverflow.com/questions/34831131/pyserial-does-not-play-well-with-virtual-port
        self.ser = serial.Serial(pty, interCharTimeout=1, rtscts=True, dsrdtr=True)

    def close(self):
        import signal

        os.killpg(os.getpgid(self.subp.pid), signal.SIGTERM)

    def read(self, size=1):
        return self.ser.read(size)

    def write(self, data):
        return self.ser.write(data)

    @property
    def in_waiting(self):
        return self.ser.in_waiting


class Pyboard:
    def __init__(
        self,
        device: Union[None, str, UsbSpecifier] = None,
        baudrate: int = 115200,
        user: str = "micro",
        password: str = "python",  # noqa: S107
        attempts: int = 1,
        exclusive: bool = True,
    ):
        """Micropython REPL compatibility class.

        Parameters
        ----------
        device: str
            Some device specificier like ``'/dev/ttyACM0'`` or ``'192.168.1.1'``.
        baudrate: int
            If a serial-like connection, this baudrate will be used.
        user: str
            If connection requires a username, this will be used.
        password: str
            If connection requires a password, this will be used.
        attempts: int
            Number of attempts to try and connect to board.
            If a ``<0`` value is provided, will infinitely try to connect.
        exclusive: bool
            If a serial-like connection, this configures the ``exclusive`` flag.
        """
        if attempts == 0:
            raise ValueError('"attempts" cannot be 0.')
        self.in_raw_repl = False
        self.use_raw_paste = True
        self._consumed_buf = bytearray()
        self._unconsumed_buf = bytearray()

        if device is None:
            usb_specifier_str = os.environ.get("BELAY_DEVICE", "{}")
            device = UsbSpecifier.parse_raw(usb_specifier_str)

        if isinstance(device, UsbSpecifier):
            device = device.to_port()
        else:
            with contextlib.suppress(ValidationError):
                device = UsbSpecifier.parse_raw(device).to_port()

        if device.startswith("exec:"):
            self.serial = ProcessToSerial(device[len("exec:") :])
        elif device.startswith("execpty:"):
            self.serial = ProcessPtyToTerminal(device[len("qemupty:") :])
        elif device and device[0].isdigit() and device[-1].isdigit() and device.count(".") == 3:
            # device looks like an IP address
            self.serial = TelnetToSerial(device, user, password, read_timeout=10)
        elif device and device.startswith("ws://"):
            self.serial = WebreplToSerial(device, password, read_timeout=10)
        else:
            import serial

            # Set options, and exclusive if pyserial supports it
            serial_kwargs = {"baudrate": baudrate, "interCharTimeout": 1}
            if serial.__version__ >= "3.3":
                serial_kwargs["exclusive"] = exclusive

            for attempt_count in itertools.count(start=1):
                try:
                    self.serial = serial.Serial(device, **serial_kwargs)
                    break
                except OSError:
                    pass

                if attempt_count == attempts:
                    raise ConnectionFailedError

                time.sleep(1.0)

        atexit.register(self.close)

    def close(self):
        if not self.serial:
            return
        self.exit_raw_repl()
        self.serial.close()
        self.serial = None
        atexit.unregister(self.close)

    def read_until(self, ending, timeout=10, data_consumer=None):
        """Read bytes until a specified ending pattern is reached.

        warning: in Belay, ``data_consumer`` may raise an exception; so make sure
        internal buffers are correct prior to calling ``data_consumer``.

        Parameters
        ----------
        data_consumer: Callable
            Function is called with data as soon as it becomes available.
            Does NOT wait for a newline.
        timeout: Union[None, float]
            Timeout in seconds.
            If None, no timeout.

        Returns
        -------
        data: bytes
            Data read up to, and including, ``ending``.
        """
        if data_consumer is None:
            data_consumer = _dummy_data_consumer

        if timeout is None:
            timeout = float("inf")
        deadline = time.time() + timeout

        def find(buf):
            # slice up to this index
            index = buf.find(ending)
            if index >= 0:
                index += len(ending)
            return index

        while True:  # loop until ``ending`` is found, or timeout
            ending_index = find(self._unconsumed_buf)
            if ending_index > 0:
                data_for_consumer = self._unconsumed_buf[:ending_index]
                self._unconsumed_buf[:] = self._unconsumed_buf[ending_index:]
                out = self._consumed_buf + data_for_consumer
                self._consumed_buf.clear()
                data_consumer(data_for_consumer)
                return out
            elif self._unconsumed_buf:
                # consume the unconsumed buffer
                og_consumed_buf_len = len(self._consumed_buf)
                self._consumed_buf.extend(self._unconsumed_buf)

                if (ending_index := find(self._consumed_buf)) > 0:
                    # The ``ending`` was split across the buffers
                    out = self._consumed_buf[:ending_index]
                    self._unconsumed_buf[:] = self._consumed_buf[ending_index:]
                    data_for_consumer = self._consumed_buf[og_consumed_buf_len:ending_index]
                    self._consumed_buf.clear()
                    data_consumer(data_for_consumer)
                    return out

                # ``ending`` has still not been found.
                data_for_consumer = self._unconsumed_buf.copy()
                self._unconsumed_buf.clear()
                data_consumer(data_for_consumer)

            while not self._unconsumed_buf:
                # loop until new data has arrived.
                if time.time() > deadline:
                    raise PyboardError(
                        f"Timed out reading until {repr(ending)}\n    Received: {repr(self._consumed_buf)}"
                    )

                if not self.serial.in_waiting:
                    time.sleep(0.001)

                if self.serial.in_waiting:
                    n_bytes = min(2048, self.serial.in_waiting)
                    self._unconsumed_buf.extend(self.serial.read(n_bytes))

    def cancel_running_program(self):
        """Interrupts any running program."""
        self.serial.write(b"\r\x03\x03")  # ctrl-C twice: interrupt any running program

    def ctrl_d(self):
        # Note: When in raw repl mode, a soft-reset will NOT execute ``main.py``.
        self.serial.write(b"\x04")  # ctrl-D: soft reset

    def enter_raw_repl(self, soft_reset=True):
        # flush input (without relying on serial.flushInput())
        n = self.serial.in_waiting
        while n > 0:
            self.serial.read(n)
            n = self.serial.in_waiting
        self.cancel_running_program()
        self.exit_raw_repl()  # if device is already in raw_repl, b'>>>' won't be printed.
        self.read_until(b">>>")
        self.serial.write(b"\r\x01")  # ctrl-A: enter raw REPL
        if soft_reset:
            self.read_until(b"raw REPL; CTRL-B to exit\r\n>")
            self.ctrl_d()

        self.read_until(b"raw REPL; CTRL-B to exit\r\n")
        self.in_raw_repl = True

    def exit_raw_repl(self):
        self.serial.write(b"\r\x02")  # ctrl-B: enter friendly REPL
        self.in_raw_repl = False

    def follow(self, timeout, data_consumer=None):
        # wait for normal output (first EOF reception)
        data = self.read_until(b"\x04", timeout=timeout, data_consumer=data_consumer)
        data = data[:-1]

        # wait for error output
        data_err = self.read_until(b"\x04", timeout=timeout)
        data_err = data_err[:-1]

        # return normal and error output
        return data, data_err

    def raw_paste_write(self, command_bytes):
        # Read initial header, with window size.
        data = self.serial.read(2)
        window_size = data[0] | data[1] << 8
        window_remain = window_size

        # Write out the command_bytes data.
        i = 0
        while i < len(command_bytes):
            while window_remain == 0 or self.serial.in_waiting:
                data = self.serial.read(1)
                if data == b"\x01":
                    # Device indicated that a new window of data can be sent.
                    window_remain += window_size
                elif data == b"\x04":
                    # Device indicated abrupt end.  Acknowledge it and finish.
                    self.serial.write(b"\x04")
                    return
                else:
                    # Unexpected data from device.
                    raise PyboardError(f"unexpected read during raw paste: {data}")
            # Send out as much data as possible that fits within the allowed window.
            b = command_bytes[i : min(i + window_remain, len(command_bytes))]
            self.serial.write(b)
            window_remain -= len(b)
            i += len(b)

        # Indicate end of data.
        self.serial.write(b"\x04")

        # Wait for device to acknowledge end of data.
        self.read_until(b"\x04")

    def exec_raw_no_follow(self, command):
        command_bytes = command if isinstance(command, bytes) else bytes(command, encoding="utf8")

        # check we have a prompt
        self.read_until(b">")

        if self.use_raw_paste:
            # Try to enter raw-paste mode.
            self.serial.write(b"\x05A\x01")
            data = self.serial.read(2)
            if data == b"R\x00":
                # Device understood raw-paste command but doesn't support it.
                pass
            elif data == b"R\x01":
                # Device supports raw-paste mode, write out the command using this mode.
                return self.raw_paste_write(command_bytes)
            else:
                # Device doesn't support raw-paste, fall back to normal raw REPL.
                self.read_until(b"w REPL; CTRL-B to exit\r\n>")
            # Don't try to use raw-paste mode again for this connection.
            self.use_raw_paste = False

        # Write command using standard raw REPL, 256 bytes every 10ms.
        for i in range(0, len(command_bytes), 256):
            self.serial.write(command_bytes[i : min(i + 256, len(command_bytes))])
            time.sleep(0.01)
        self.serial.write(b"\x04")

        # check if we could exec command
        data = self.serial.read(2)
        if data != b"OK":
            raise PyboardError("could not exec command (response: %r)" % data)

    def exec_raw(self, command, timeout=None, data_consumer=None):
        self.exec_raw_no_follow(command)
        return self.follow(timeout, data_consumer)

    def eval(self, expression):
        ret = self.exec(f"print({expression})")
        ret = ret.strip()
        return ret

    def exec(self, command, data_consumer=None):
        ret, ret_err = self.exec_raw(command, data_consumer=data_consumer)
        if ret_err:
            raise PyboardException(ret_err.decode())
        return ret

    def execfile(self, filename):
        pyfile = Path(filename).read_bytes()
        return self.exec(pyfile)

    def get_time(self):
        t = str(self.eval("pyb.RTC().datetime()"), encoding="utf8")[1:-1].split(", ")
        return int(t[4]) * 3600 + int(t[5]) * 60 + int(t[6])

    def fs_ls(self, src):
        cmd = (
            "import uos\nfor f in uos.ilistdir(%s):\n"
            " print('{:12} {}{}'.format(f[3]if len(f)>3 else 0,f[0],'/'if f[1]&0x4000 else ''))"
            % (("'%s'" % src) if src else "")
        )
        self.exec(cmd, data_consumer=stdout_write_bytes)

    def fs_cat(self, src, chunk_size=256):
        cmd = "with open('%s') as f:\n while 1:\n  b=f.read(%u)\n  if not b:break\n  print(b,end='')" % (
            src,
            chunk_size,
        )
        self.exec(cmd, data_consumer=stdout_write_bytes)

    def fs_get(self, src, dest, chunk_size=256, progress_callback=None):
        dest = Path(dest)
        written = 0
        if progress_callback:
            src_size = int(self.exec("import os\nprint(os.stat('%s')[6])" % src))
        self.exec("f=open('%s','rb')\nr=f.read" % src)
        with dest.open("wb") as f:
            while True:
                data = bytearray()
                self.exec(
                    "print(r(%u))" % chunk_size,
                    data_consumer=lambda d: data.extend(d),  # noqa: B023
                )
                if not data.endswith(b"\r\n\x04"):
                    raise PyboardError

                try:
                    data = ast.literal_eval(str(data[:-3], "ascii"))
                except Exception as e:
                    raise PyboardError from e

                if not isinstance(data, bytes):
                    raise PyboardError

                if not data:
                    break
                f.write(data)
                written += len(data)
                if progress_callback:
                    progress_callback(written, src_size)
        self.exec("f.close()")

    def fs_put(self, src, dest, chunk_size=256, progress_callback=None):
        src = Path(src)
        written = 0
        src_size = src.stat().st_size
        self.exec("f=open('%s','wb')\nw=f.write" % dest)
        with src.open("rb") as f:
            while True:
                data = f.read(chunk_size)
                if not data:
                    break
                written += len(data)
                self.exec("w(" + repr(data) + ")")
                if progress_callback:
                    progress_callback(written, src_size)
        self.exec("f.close()")

    def fs_mkdir(self, dir):
        self.exec("import uos\nuos.mkdir('%s')" % dir)

    def fs_rmdir(self, dir):
        self.exec("import uos\nuos.rmdir('%s')" % dir)

    def fs_rm(self, src):
        self.exec("import uos\nuos.remove('%s')" % src)



from pathlib import Path
from typing import Union


def fnv1a(fn: Union[str, Path]) -> int:
    """Compute the FNV-1a 32-bit hash of a file."""
    fn = Path(fn)
    h = 0x811C9DC5
    size = 1 << 32
    with fn.open("rb") as f:
        while True:
            data = f.read(65536)
            if not data:
                break
            for byte in data:
                h = h ^ byte
                h = (h * 0x01000193) % size
    return h



import tokenize
from collections import deque
from io import StringIO
from tokenize import COMMENT, DEDENT, INDENT, NEWLINE, OP, STRING

_token_type_line_start = {NEWLINE, DEDENT, INDENT}


def minify(code: str) -> str:
    """Minify python code.

    Naive code minifying that preserves names and linenos. Performs the following:

        * Removes docstrings.

        * Removes comments.

        * Removes unnecessary whitespace.

    Parameters
    ----------
    code: str
        Python code to minify.

    Returns
    -------
    str
        Minified code.
    """
    out = []
    last_lineno = -1
    last_col = 0
    prev_start_line = 0
    level = 0
    global_start_col = 0
    prev_token_types = deque([INDENT], maxlen=2)

    for (
        token_type,
        string,
        (start_line, start_col),
        (end_line, end_col),
        _,
    ) in tokenize.generate_tokens(StringIO(code).readline):
        prev_token_types.append(token_type)
        prev_token_type = prev_token_types.popleft()

        if start_line > last_lineno:
            last_col = global_start_col

        if token_type == INDENT:
            if start_line == 1:
                # Whole code-block is indented.
                global_start_col = end_col
            else:
                level += 1
            continue
        elif token_type == DEDENT:
            level -= 1
            continue
        elif token_type == COMMENT:
            continue

        if token_type == STRING and (prev_token_type in (NEWLINE, INDENT) or start_col == global_start_col):
            # Docstring
            out.append(" " * level + "0" + "\n" * string.count("\n"))
        elif start_line > prev_start_line and token_type != NEWLINE:
            # First op of a line, needs its minimized indent
            out.append(" " * level)  # Leading indent
            if string == "pass":
                out.append("0")  # Shorter than a ``pass`` statement.
            else:
                out.append(string)
        elif token_type == OP and prev_token_type not in _token_type_line_start:
            # No need for a space before operators.
            out.append(string)
        elif start_col > last_col and token_type != NEWLINE:
            if prev_token_type != OP:
                out.append(" ")
            out.append(string)
        else:
            out.append(string)

        prev_start_line = start_line
        last_col = end_col
        last_lineno = end_line

    return "".join(out)



import shutil

from belay.project import find_dependencies_folder, load_groups


def clean():
    """Remove any downloaded dependencies if they are no longer specified in pyproject."""
    groups = load_groups()
    dependencies_folder = find_dependencies_folder()

    existing_group_folders = {x for x in dependencies_folder.glob("*") if x.is_dir()}

    # Remove missing dependencies in each group
    for group in groups:
        group.clean()
        existing_group_folders.discard(group.folder)

    # Remove missing group folders
    for group_folder in existing_group_folders:
        shutil.rmtree(group_folder)



# Don't manually change, let poetry-dynamic-versioning-plugin handle it.
__version__ = "0.21.0"

__all__ = [
    "AuthenticationError",
    "ConnectionFailedError",
    "ConnectionLost",
    "Device",
    "DeviceMeta",
    "DeviceNotFoundError",
    "FeatureUnavailableError",
    "Implementation",
    "InsufficientSpecifierError",
    "MaxHistoryLengthError",
    "NoMatchingExecuterError",
    "NotBelayResponseError",
    "PyboardException",
    "SpecialFunctionNameError",
    "UsbSpecifier",
    "list_devices",
    "minify",
]
from ._minify import minify
from .device import Device, Implementation
from .device_meta import DeviceMeta
from .exceptions import (
    AuthenticationError,
    ConnectionFailedError,
    ConnectionLost,
    DeviceNotFoundError,
    FeatureUnavailableError,
    InsufficientSpecifierError,
    MaxHistoryLengthError,
    NoMatchingExecuterError,
    NotBelayResponseError,
    SpecialFunctionNameError,
)
from .pyboard import PyboardException
from .usb_specifier import UsbSpecifier, list_devices



import ast
import inspect
import re
from io import StringIO
from tokenize import (
    COMMENT,
    DEDENT,
    INDENT,
    NEWLINE,
    OP,
    STRING,
    generate_tokens,
    untokenize,
)
from typing import Tuple

_pat_no_decorators = re.compile(r"^(\s*def\s)|(\s*async\s+def\s)|(.*(?<!\w)lambda(:|\s))")


class _NoAction(Exception):  # noqa: N818
    pass


def _dedent_tokenizer(code):
    indent_to_remove = ""
    for (
        token_type,
        string,
        (start_line, start_col),
        (end_line, end_col),  # noqa: B007
        _,
    ) in generate_tokens(StringIO(code).readline):
        if start_line == 1 and start_col == 0:
            # First Token
            if token_type != INDENT:
                # No action to perform
                raise _NoAction
            indent_to_remove = string
        if start_col == 0 and token_type == INDENT:
            if not string.startswith(indent_to_remove):
                raise IndentationError
            string = string[len(indent_to_remove) :]
        yield token_type, string


def _dedent(code):
    try:
        return untokenize(_dedent_tokenizer(code))
    except _NoAction:
        return code


def _remove_signature(code):
    tree = ast.parse(code)
    lines = code.split("\n")
    lines_removed = tree.body[0].body[0].lineno - 1
    # TODO: won't properly handle single line functions.
    # Or other weirdly formatted stuff.
    lines = lines[lines_removed:]
    return "\n".join(lines), lines_removed


def getsource(f, *, strip_signature=False) -> Tuple[str, int, str]:
    """Get source code with mild post processing.

    * strips leading decorators.
    * Trims leading whitespace indent.

    Parameters
    ----------
    f: Callable
        Function to get source code of.
    strip_signature: bool
        Remove the function signature from the returned code.

    Returns
    -------
    src_code: str
        Source code. Always ends in a newline.
    src_lineno: int
        Line number of code begin.
    src_file: str
        Path to file containing source code.
    """
    src_file = inspect.getsourcefile(f)
    if src_file is None:
        raise FileNotFoundError(f"Unable to get source file for {f}.")
    lines, src_lineno = inspect.getsourcelines(f)

    offset = 0
    for line in lines:
        if _pat_no_decorators.match(line):
            break
        offset += 1

    lines = lines[offset:]

    src_code = "".join(lines)
    src_lineno += offset

    src_code = _dedent(src_code)

    if strip_signature:
        src_code, lines_removed = _remove_signature(src_code)
        src_lineno += lines_removed

        src_code = _dedent(src_code)

    return src_code, src_lineno, src_file


def isexpression(code: str) -> bool:
    """Checks if python code is an expression.

    Parameters
    ----------
    code: str
        Python code to check if is an expression.

    Returns
    -------
    bool
        ``True`` if ``code`` is an expression; returns
        ``False`` otherwise (statement or invalid).
    """
    try:
        compile(code, "<stdin>", "eval")
    except SyntaxError:
        return False

    return True



import os
import shutil
import subprocess  # nosec
import sys
from tempfile import TemporaryDirectory
from typing import List

import typer
from typer import Option

import belay
from belay.cli import cache
from belay.cli.clean import clean
from belay.cli.exec import exec
from belay.cli.info import info
from belay.cli.install import install
from belay.cli.new import new
from belay.cli.run import run
from belay.cli.select import select
from belay.cli.sync import sync
from belay.cli.terminal import terminal
from belay.cli.update import update
from belay.project import load_groups

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)
app.add_typer(cache.app, name="cache")

app.command()(clean)
app.command()(exec)
app.command()(info)
app.command()(install)
app.command()(new)
app.command()(run)
app.command()(select)
app.command()(sync)
app.command()(terminal)
app.command()(update)


def run_exec(command: List[str]):
    """Enable virtual-environment and run command."""
    groups = load_groups()
    virtual_env = os.environ.copy()
    # Add all dependency groups to the micropython path.
    # This flattens all dependencies to a single folder and fetches fresh
    # copies of dependencies in ``develop`` mode.
    with TemporaryDirectory() as tmp_dir:
        virtual_env["MICROPYPATH"] = f".:{tmp_dir}"
        for group in groups:
            group.copy_to(tmp_dir)

        try:
            subprocess.run(  # nosec
                command,
                env=virtual_env,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            sys.exit(e.returncode)


def _get(indexable, index, default=None):
    try:
        return indexable[index]
    except IndexError:
        return default


def run_app(*args, **kwargs):
    """Add CLI hacks that are not Typer-friendly here."""
    command = _get(sys.argv, 1)

    try:
        exec_path = shutil.which(sys.argv[2])
    except IndexError:
        exec_path = None

    if command == "run" and exec_path:
        # Special subcommand override.
        run_exec(sys.argv[2:])
    else:
        # Common-case; use Typer functionality.
        app(*args, **kwargs)


def version_callback(value: bool):
    if not value:
        return
    print(belay.__version__)
    raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = Option(
        None,
        "--version",
        "-v",
        callback=version_callback,
        help="Display Belay's version.",
    ),
):
    pass



"""Metaclass to allow executer overloading.

Inspired by mCoding example::

    https://github.com/mCodingLLC/VideosSampleCode/blob/master/videos/077_metaclasses_in_python/overloading.py
"""
from autoregistry import RegistryMeta

from .exceptions import NoMatchingExecuterError

_MISSING = object()


class OverloadList(list):
    """To separate user-lists from a list of overloaded methods."""


class OverloadDict(dict):
    """Dictionary where a key can be written to multiple times."""

    def __setitem__(self, key, value):
        # Key: method name
        # Value: method

        if not isinstance(key, str):
            raise TypeError

        prior_val = self.get(key, _MISSING)
        method_metadata = getattr(value, "__belay__", False)

        if prior_val is _MISSING:
            # Register a new method name
            insert_val = OverloadList([value]) if method_metadata else value
            super().__setitem__(key, insert_val)
        elif isinstance(prior_val, OverloadList):
            # Add to a previously overloaded method.
            if not method_metadata:
                raise ValueError(f"Cannot mix non-executer and executer methods: {key}")

            # Check for a previous "catchall" method
            for f in prior_val:
                if not f.__belay__.implementation:
                    raise ValueError(f"Cannot define another executor after catchall: {key}.")
            prior_val.append(value)
        else:
            # Overwrite a previous vanilla method
            if method_metadata:
                raise ValueError(f"Cannot mix non-executer and executer methods: {key}")
            super().__setitem__(key, value)


class ExecuterMethod:
    def __set_name__(self, owner, name):
        self.owner = owner
        self.name = name

    def __init__(self, overload_list):
        if not isinstance(overload_list, OverloadList):
            raise TypeError("Must use OverloadList.")
        if not overload_list:
            raise ValueError("Empty OverloadList.")
        self.overload_list = overload_list

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.overload_list!r})"

    def __get__(self, instance, _owner=None):
        if instance is None:
            return self
        # Don't use owner == type(instance)
        # We want self.owner, which is the class from which get is being called
        for f in self.overload_list:
            imp = f.__belay__.implementation

            if not imp or imp == instance.implementation.name:
                return f

        # no matching overload in owner class, check next in line
        super_instance = super(self.owner, instance)
        super_call = getattr(super_instance, self.name, _MISSING)
        if super_call is not _MISSING:
            return super_call
        else:
            raise NoMatchingExecuterError()


class DeviceMeta(RegistryMeta):
    @classmethod
    def __prepare__(cls, name, bases, **kwargs):
        return OverloadDict()

    def __new__(cls, name, bases, namespace, **kwargs):
        overload_namespace = {
            key: ExecuterMethod(val) if isinstance(val, OverloadList) else val for key, val in namespace.items()
        }
        output_cls = super().__new__(cls, name, bases, overload_namespace, **kwargs)
        return output_cls

    def mro(self):
        mro = super().mro()
        # Move ``Device`` to back, if it exists in the mro

        try:
            from belay.device import Device
        except ImportError:  # circular import on first use
            return mro

        try:
            mro.remove(Device)
        except ValueError:  # Device was not in ``bases``
            return mro

        # Insert it right before ``object``
        mro.insert(mro.index(object), Device)

        return mro



import shutil
from pathlib import Path
from urllib.parse import urlparse

import fsspec
from autoregistry import Registry

from belay.typing import PathType

# Downloaders should have function signature
#    def downloader(dst: Path, uri: str) -> Path
# where the return value is one of:
#    1. ``dst`` if a folder was downloaded
#    2. Path to single file if the URI was for a single file.
downloaders = Registry()


class NonMatchingURI(Exception):  # noqa: N818
    """Provided URI does not match downloading function."""


# DO NOT decorate with ``@downloaders``, since this must be last.
def _download_generic(dst: Path, uri: str) -> Path:
    """Downloads a single file or folder to ``dst / <filename>``."""
    parsed = urlparse(uri)

    if parsed.scheme in ("", "file"):
        # Local file, make it relative to project root
        uri_path = Path(uri)

        if not uri_path.is_absolute():
            from belay.project import find_project_folder

            uri_path = find_project_folder() / uri

        uri = str(uri_path)

    if Path(uri).is_dir():  # local
        shutil.copytree(uri, dst, dirs_exist_ok=True)
    else:
        with fsspec.open(uri, "rb") as f:
            data = f.read()

        dst /= Path(uri).name
        with dst.open("wb") as f:
            f.write(data)

    return dst


def download_uri(dst_folder: PathType, uri: str) -> Path:
    """Download ``uri`` by trying all downloaders on ``uri`` until one works."""
    dst_folder = Path(dst_folder)
    for processor in downloaders.values():
        try:
            return processor(dst_folder, uri)
            break
        except NonMatchingURI:
            pass
    else:
        return _download_generic(dst_folder, uri)



import contextlib
import hashlib
import shutil
from pathlib import Path

from belay.typing import PathType


def _sha256sum(path: PathType):
    path = Path(path)
    h = hashlib.sha256()
    mv = memoryview(bytearray(128 * 1024))
    with path.open("rb", buffering=0) as f:
        while n := f.readinto(mv):
            h.update(mv[:n])
    return h.hexdigest()


def sync(src_folder: PathType, dst_folder: PathType) -> bool:
    """Make ``dst_folder`` have the same contents as ``src_folder``.

    Returns
    -------
    bool
        ``True`` if contents of ``dst`` have changed; ``False`` otherwise.
    """
    changed = False
    src_folder, dst_folder = Path(src_folder), Path(dst_folder)

    src_files = {x.relative_to(src_folder) for x in src_folder.rglob("*") if x.is_file()}

    dst_files, dst_subfolders = set(), set()
    for dst_item in dst_folder.rglob("*"):
        dst_item_rel = dst_item.relative_to(dst_folder)
        if dst_item.is_file():
            dst_files.add(dst_item_rel)
        else:
            dst_subfolders.add(dst_item_rel)

    common_files = src_files.intersection(dst_files)
    src_only_files = src_files - dst_files
    dst_only_files = dst_files - src_files

    # compare common files and copy over on change
    for f in common_files:
        src = src_folder / f
        dst = dst_folder / f

        if _sha256sum(src) != _sha256sum(dst):
            changed = True
            shutil.copy(src, dst)

    # copy over src_only_files
    for f in src_only_files:
        changed = True
        src = src_folder / f
        dst = dst_folder / f
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(src, dst)

    # Remove files that only exist in the destination
    for f in dst_only_files:
        changed = True
        dst = dst_folder / f
        dst.unlink()

    # Remove all empty dst directories.
    for folder in dst_subfolders:
        with contextlib.suppress(OSError):
            folder.rmdir()

    return changed



from pathlib import Path
from typing import Callable, Dict, Generator, List, Set, Union

PythonLiteral = Union[None, bool, bytes, int, float, str, List, Dict, Set]
BelayGenerator = Generator[PythonLiteral, None, None]
BelayReturn = Union[BelayGenerator, PythonLiteral]
BelayCallable = Callable[..., BelayReturn]

PathType = Union[str, Path]



from typer import Argument, Option

from belay import Device
from belay.cli.common import help_password, help_port


def terminal(
    port: str = Argument(..., help=help_port),
    password: str = Option("", help=help_password),
):
    """Open up an interactive REPL.

    Press ctrl+] to exit.
    """
    with Device(port, password=password) as device:
        device.terminal()



import ast
import atexit
import concurrent.futures
import contextlib
import importlib.resources
import linecache
import re
import shutil
import sys
from inspect import signature
from pathlib import Path
from tempfile import TemporaryDirectory
from types import ModuleType
from typing import Callable, Optional, TextIO, TypeVar, Union, overload

from serial import SerialException
from serial.tools.miniterm import Miniterm
from typing_extensions import ParamSpec

from ._minify import minify as minify_code
from .device_meta import DeviceMeta
from .device_support import Implementation, MethodMetadata, sort_executers
from .device_sync_support import (
    discover_files_dirs,
    generate_dst_dirs,
    preprocess_ignore,
    preprocess_keep,
    preprocess_src_file_hash,
)
from .exceptions import (
    ConnectionLost,
    InternalError,
    MaxHistoryLengthError,
    NotBelayResponseError,
)
from .executers import (
    Executer,
    SetupExecuter,
    TaskExecuter,
    TeardownExecuter,
    ThreadExecuter,
)
from .helpers import read_snippet, wraps_partial
from .inspect import isexpression
from .pyboard import Pyboard, PyboardError, PyboardException
from .typing import BelayReturn, PathType
from .webrepl import WebreplToSerial

P = ParamSpec("P")
R = TypeVar("R")


def parse_belay_response(line):
    if not line.startswith("_BELAY"):
        raise NotBelayResponseError
    line = line[6:]
    code, line = line[0], line[1:]

    if code == "R":
        # Result
        return ast.literal_eval(line)
    elif code == "S":
        # StopIteration
        raise StopIteration
    else:
        raise ValueError(f'Received unknown code: "{code}"')


class Device(metaclass=DeviceMeta):
    """Belay interface into a micropython device.

    Can be used as a context manager; calls ``self.close`` on exit.

    Uses the ``autoregistry.RegistryMeta`` metaclass for easy-to-access subclasses.

    Attributes
    ----------
    implementation: Implementation
        Implementation details of device.
    """

    MAX_CMD_HISTORY_LEN = 1000

    def __init__(
        self,
        *args,
        startup: Optional[str] = None,
        attempts: int = 0,
        **kwargs,
    ):
        """Create a MicroPython device.

        Parameters
        ----------
        startup: str
            Code to run on startup. Defaults to a few common imports.
        attempts: int
            If device disconnects, attempt to re-connect this many times (with 1 second between attempts).
            WARNING: this may result in unexpectedly long blocking calls when reconnecting!
        """
        self._board_kwargs = signature(Pyboard).bind(*args, **kwargs).arguments
        self.attempts = attempts
        self._cmd_history = []

        self._connect_to_board(**self._board_kwargs)

        self._exec_snippet("startup")

        # Obtain implementation early on so implementation-specific executers can be bound.
        self.implementation = Implementation(
            *self("(sys.implementation.name, sys.implementation.version, sys.platform)"),
            emitters=self._emitter_check(),
        )

        # Setup executer generators and bind to private attributes.
        executer_generators = {}
        for executer_name, executer_cls in Executer.items():
            executer_generator = executer_cls(self)
            # Private interface, will always be available
            attr_name = f"_belay_{executer_name}"
            setattr(self, attr_name, executer_generator)
            executer_generators[executer_name] = executer_generator

        # If subclassing Device, register methods decorated with
        # executer markers (e.g. ``@Device.task``).
        autoinit_executers = []
        instantiated_executer_names = set()
        for method_name in dir(type(self)):
            # Get method from self to trigger descriptors.
            try:
                method = getattr(self, method_name)
                metadata = method.__belay__
            except AttributeError:
                continue
            executer_name = metadata.executer.__registry__.name
            executer_generator = executer_generators[executer_name]
            executer = executer_generator(method, **metadata.kwargs)
            instantiated_executer_names.add(executer_name)

            if metadata.autoinit:
                autoinit_executers.append(executer)

            setattr(
                self,
                method_name,
                executer,
            )

        # Setup publicly accessible if the name hasn't been stomped.
        for executer_name, executer_generator in executer_generators.items():
            if executer_name in instantiated_executer_names:
                continue
            setattr(self, executer_name, executer_generator)

        if startup is None:
            if self.implementation.name == "circuitpython":
                self._exec_snippet("convenience_imports_circuitpython")
            else:
                self._exec_snippet("convenience_imports_micropython")
        elif startup:
            self(startup)

        self.__pre_autoinit__()

        for executer in sort_executers(autoinit_executers):
            executer()

        atexit.register(self.close)

        self.__post_init__()

    def __pre_autoinit__(self):
        """Runs near the end of ``__init__``, but before methods marked with ``setup(autoinit=True)`` are invoked.

        This would be a good location to call items like:
            * ``self.sync(...)`` - Basic file sync
            * ``self.sync_dependencies(...)`` - More advanced sync.
              Recommended way of getting dependencies on-device.
        """
        pass

    def __post_init__(self):
        """Runs at the very end of ``__init__``."""
        pass

    def _emitter_check(self):
        # Detect which emitters are available
        emitters = []
        try:
            self._exec_snippet("emitter_check")
        except PyboardException as e:
            if "invalid micropython decorator" not in str(e):
                raise
            # Get line of exception
            line_e = int(re.findall(r"line (\d+)", str(e))[-1])
            if line_e == 1:
                # No emitters available
                pass
            else:
                emitters.append("native")
                if line_e == 3:
                    # viper is not available
                    pass
                else:
                    raise InternalError(f"Unknown emitter line {line_e}.") from e
        else:
            emitters.append("native")
            emitters.append("viper")
            self("del __belay_emitter_test")

        return tuple(emitters)

    def _connect_to_board(self, **kwargs):
        self._board = Pyboard(**kwargs)
        soft_reset = not isinstance(self._board.serial, WebreplToSerial)
        self._board.enter_raw_repl(soft_reset=soft_reset)

    def _exec_snippet(self, *names: str) -> BelayReturn:
        """Load and execute a snippet from the snippets sub-package.

        Parameters
        ----------
        names : str
            Snippet(s) to load and execute.
        """
        snippets = [read_snippet(name) for name in names]
        return self("\n".join(snippets))

    def __call__(
        self,
        cmd: str,
        *,
        minify: bool = True,
        stream_out: TextIO = sys.stdout,
        record=True,
    ):
        """Execute code on-device.

        Parameters
        ----------
        cmd: str
            Python code to execute. May be a statement or expression.
        minify: bool
            Minify ``cmd`` code prior to sending.
            Reduces the number of characters that need to be transmitted.
            Defaults to ``True``.
        record: bool
            Record the call for state-reconstruction if device is accidentally reset.
            Defaults to ``True``.

        Returns
        -------
            Correctly interpreted return value from executing code on-device.
        """
        if minify:
            cmd = minify_code(cmd)

        if isexpression(cmd):
            # Belay Tasks are inherently expressions as well.
            cmd = f"print('_BELAYR' + repr({cmd}))"

        if record and self.attempts and len(self._cmd_history) < self.MAX_CMD_HISTORY_LEN:
            self._cmd_history.append(cmd)

        out = None  # Used to store the parsed response object.
        data_consumer_buffer = bytearray()

        def data_consumer(data):
            """Handle input data stream immediately."""
            nonlocal out
            data = data.replace(b"\x04", b"")
            if not data:
                return
            data_consumer_buffer.extend(data)
            while (i := data_consumer_buffer.find(b"\n")) >= 0:
                i += 1
                line = data_consumer_buffer[:i].decode()
                data_consumer_buffer[:] = data_consumer_buffer[i:]
                try:
                    out = parse_belay_response(line)
                except NotBelayResponseError:
                    if stream_out:
                        stream_out.write(line)

        try:
            self._board.exec(cmd, data_consumer=data_consumer)
        except (SerialException, ConnectionResetError) as e:
            # Board probably disconnected.
            if self.attempts:
                self.reconnect()
                self._board.exec(cmd, data_consumer=data_consumer)
            else:
                raise ConnectionLost from e

        return out

    def sync(
        self,
        folder: PathType,
        dst: str = "/",
        keep: Union[None, list, str, bool] = None,
        ignore: Union[None, list, str] = None,
        minify: bool = True,
        mpy_cross_binary: Union[str, Path, None] = None,
        progress_update=None,
    ) -> None:
        """Sync a local directory to the remote filesystem.

        For each local file, check the remote file's hash, and transfer if they differ.
        If a file/folder exists on the remote filesystem that doesn't exist in the local
        folder, then delete it (unless it's in ``keep``).

        Parameters
        ----------
        folder: str, Path
            Single file or directory of files to sync to the root of the board's filesystem.
        dst: str
            Destination **directory** on device.
            Defaults to unpacking ``folder`` to root.
        keep: None | str | list | bool
            Do NOT delete these file(s) on-device if not present in ``folder``.
            If ``true``, don't delete any files on device.
            If ``false``, delete all unsynced files (same as passing ``[]``).
            If ``dst is None``, defaults to ``["boot.py", "webrepl_cfg.py", "lib"]``.
        ignore: None | str | list
            Git's wildmatch patterns to NOT sync to the device.
            Defaults to ``["*.pyc", "__pycache__", ".DS_Store", ".pytest_cache"]``.
        minify: bool
            Minify python files prior to syncing.
            Defaults to ``True``.
        mpy_cross_binary: Union[str, Path, None]
            Path to mpy-cross binary. If provided, ``.py`` will automatically
            be compiled.
            Takes precedence over minifying.
        progress_update:
            Partial for ``rich.progress.Progress.update(task_id,...)`` to update with sync status.
        """
        folder = Path(folder).resolve()

        dst = str(dst)
        if not dst.startswith("/"):
            raise ValueError('dst must start with "/"')
        elif len(dst) > 1:
            dst = dst.rstrip("/")

        if not folder.exists():
            raise ValueError(f'"{folder}" does not exist.')

        # Create a list of all files and dirs (on-device).
        # This is so we know what to clean up after done syncing.
        snippets_to_execute = []

        if progress_update:
            progress_update(description="Bootstrapping sync...")
        if "viper" in self.implementation.emitters:
            snippets_to_execute.append("hf_viper")
        elif "native" in self.implementation.emitters:
            snippets_to_execute.append("hf_native")
        else:
            snippets_to_execute.append("hf")

        if self.implementation.name == "circuitpython":
            snippets_to_execute.append("ilistdir_circuitpython")
        else:
            snippets_to_execute.append("ilistdir_micropython")
        snippets_to_execute.append("sync_begin")
        self._exec_snippet(*snippets_to_execute)

        # Remove the keep files from the on-device ``all_files`` set
        # so they don't get deleted.
        keep_all = folder.is_file() or keep is True
        keep = preprocess_keep(keep, dst)
        ignore = preprocess_ignore(ignore)

        src_files, src_dirs, dst_files = discover_files_dirs(dst, folder, ignore)

        if mpy_cross_binary:
            dst_files = [
                dst_file.with_suffix(".mpy") if dst_file.suffix == ".py" else dst_file for dst_file in dst_files
            ]
        dst_files = [dst_file.as_posix() for dst_file in dst_files]
        dst_dirs = generate_dst_dirs(dst, folder, src_dirs)

        if keep_all:
            self("del __belay_del_fs")
        else:
            self(f"__belay_del_fs({repr(dst)}, {repr(set(keep + dst_files))}); del __belay_del_fs")

        # Try and make all remote dirs
        if dst_dirs:
            if progress_update:
                progress_update(description="Creating remote directories...")
            self(f"__belay_mkdirs({repr(dst_dirs)})")

        with TemporaryDirectory() as tmp_dir, concurrent.futures.ThreadPoolExecutor() as executor:
            tmp_dir = Path(tmp_dir)

            def _preprocess_src_file_hash_helper(src_file):
                return preprocess_src_file_hash(tmp_dir, src_file, minify, mpy_cross_binary)

            src_files_and_hashes = executor.map(_preprocess_src_file_hash_helper, src_files)

            # Get all remote hashes
            if progress_update:
                progress_update(description="Fetching remote hashes...")
            dst_hashes = self(f"__belay_hfs({repr(dst_files)})")

            if len(dst_hashes) != len(dst_files):
                raise InternalError

            puts = []
            for (src_file, src_hash), dst_file, dst_hash in zip(src_files_and_hashes, dst_files, dst_hashes):
                if src_hash != dst_hash:
                    puts.append((src_file, dst_file))

            if progress_update:
                progress_update(total=len(puts))

            for src_file, dst_file in puts:
                if progress_update:
                    progress_update(description=f"Pushing: {dst_file[1:]}")
                self._board.fs_put(src_file, dst_file)
                if progress_update:
                    progress_update(advance=1)

    def sync_dependencies(
        self,
        package: Union[ModuleType, str],
        *subfolders: PathType,
        dst="/lib",
        **kwargs,
    ):
        """Convenience method for syncing dependencies bundled with package.

        If using Belay's package manager feature, set ``dependencies_path``
        to a folder *inside* your python package (e.g.
        ``dependencies_path="mypackage/dependencies"``).

        The following example will sync all the files/folders in
        ``mypackage/dependencies/main`` to device's ``/lib``.

        .. code-block:: python

            import mypackage

            device.sync_package(mypackage, "dependencies/main")

        For intended use, ``sync_dependencies`` should be **only be called
        once**. Multiple invocations overwrite/delete previous calls' contents.

        .. code-block:: python

            # Good
            device.sync_package(mypackage, "dependencies/main", "dependencies/dev")

            # Bad (deletes on-device files from "dependencies/main")
            device.sync_package(mypackage, "dependencies/main")
            device.sync_package(mypackage, "dependencies/dev")

        Parameters
        ----------
        package: Union[ModuleType, str]
            Either the imported package or the name of a package that
            contains the data we would like to sync.
        *subfolders
            Subfolder(s) to combine and then sync to ``dst``.
            Typically something like "dependencies/main"
        dst: Union[str, Path]
            On-device destination directory.
            Defaults to ``/lib``.
        **kwargs
            Passed along to ``Device.sync``.
        """
        pkg_files = importlib.resources.files(package)
        with TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            for subfolder in subfolders:
                with importlib.resources.as_file(pkg_files / str(subfolder)) as f:
                    shutil.copytree(f, tmp_dir, dirs_exist_ok=True)
            self.sync(tmp_dir, dst=dst, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def close(self) -> None:
        """Close the connection to device.

        Automatically called on context manager exit.
        """
        # Invoke all teardown executers prior to closing out connection.
        if self._board is None:
            # Has already been closed
            return

        atexit.unregister(self.close)

        self._board.cancel_running_program()

        for executer in sort_executers(self._belay_teardown._belay_executers):
            executer()

        self._board.close()
        self._board = None

    def reconnect(self, attempts: Optional[int] = None) -> None:
        """Reconnect to the device and replay the command history.

        Parameters
        ----------
        attempts : int
            Number of times to attempt to connect to board with a 1 second delay in-between.
            If ``None``, defaults to whatever value was supplied to init.
            If init value is 0, then defaults to 1.
        """
        if len(self._cmd_history) == self.MAX_CMD_HISTORY_LEN:
            raise MaxHistoryLengthError

        kwargs = self._board_kwargs.copy()
        kwargs["attempts"] = attempts
        if kwargs["attempts"] is None:
            kwargs["attempts"] = self.attempts if self.attempts else 1

        try:
            self._connect_to_board(**kwargs)
        except PyboardError as e:
            raise ConnectionLost from e

        # Playback the history
        for cmd in self._cmd_history:
            self(cmd, record=False)

    @overload
    @staticmethod
    def setup(f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    @staticmethod
    def setup(
        *, autoinit: bool = False, implementation: str = "", **kwargs
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    @staticmethod
    def setup(
        f: Optional[Callable[P, R]] = None,
        autoinit: bool = False,
        implementation: str = "",
        **kwargs,
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        """Execute decorated function's body in a global-context on-device when called.

        Function arguments are also set in the global context.

        Can either be used as a staticmethod ``@Device.setup`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.setup`` for marking functions to a specific ``Device`` instance.

        Parameters
        ----------
        f: Callable
            Function to decorate. Can only accept and return python literals.
        minify: bool
            Minify ``cmd`` code prior to sending.
            Defaults to ``True``.
        register: bool
            Assign an attribute to ``self.setup`` with same name as ``f``.
            Defaults to ``True``.
        record: bool
            Each invocation of the executer is recorded for playback upon reconnect.
            Defaults to ``True``.
        autoinit: bool
            Automatically invokes decorated functions at the end of object ``__init__``.
            Methods will be executed in order-registered.
            Defaults to ``False``.
        ignore_errors: bool
            Discard any device-side uncaught exception.
            Defaults to ``False``.
        implementation: str
            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
            Several methods of the same name can be overloaded that support different implementations.
            Common values include "micropython", and "circuitpython".
            Defaults to an empty string, which **all** implementations will match to.
        """  # noqa: D400
        if f is None:
            return wraps_partial(Device.setup, autoinit=autoinit, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]
        if signature(f).parameters and autoinit:
            raise ValueError(f"Method {f} decorated with @Device.setup(autoinit=True) must have no arguments.")

        f.__belay__ = MethodMetadata(
            executer=SetupExecuter,
            autoinit=autoinit,
            implementation=implementation,
            kwargs=kwargs,
        )
        return f

    @overload
    @staticmethod
    def teardown(f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    @staticmethod
    def teardown(*, implementation: str = "", **kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    @staticmethod
    def teardown(
        f: Optional[Callable[P, R]] = None, implementation: str = "", **kwargs
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        """Executes decorated function's body in a global-context on-device when ``device.close()`` is called.

        Function arguments are also set in the global context.

        Can either be used as a staticmethod ``@Device.teardown`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.teardown`` for marking functions to a specific ``Device`` instance.

        Parameters
        ----------
        f: Callable
            Function to decorate. Can only accept and return python literals.
        minify: bool
            Minify ``cmd`` code prior to sending.
            Defaults to ``True``.
        register: bool
            Assign an attribute to ``self.teardown`` with same name as ``f``.
            Defaults to ``True``.
        record: bool
            Each invocation of the executer is recorded for playback upon reconnect.
            Defaults to ``True``.
        ignore_errors: bool
            Discard any device-side uncaught exception.
            Defaults to ``False``.
        implementation: str
            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
            Several methods of the same name can be overloaded that support different implementations.
            Common values include "micropython", and "circuitpython".
            Defaults to an empty string, which **all** implementations will match to.
        """  # noqa: D400
        if f is None:
            return wraps_partial(Device.teardown, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]

        if signature(f).parameters:
            raise ValueError(f'Method {f} decorated with "@Device.teardown" must have no arguments.')

        f.__belay__ = MethodMetadata(executer=TeardownExecuter, implementation=implementation, kwargs=kwargs)
        return f

    @overload
    @staticmethod
    def task(f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    @staticmethod
    def task(*, implementation: str = "", **kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    @staticmethod
    def task(
        f: Optional[Callable[P, R]] = None, implementation: str = "", **kwargs
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        """Execute decorated function on-device.

        Sends source code to device at decoration time.
        Execution sends involves much smaller overhead.

        Can either be used as a staticmethod ``@Device.task`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.task`` for marking functions to a specific ``Device`` instance.

        Parameters
        ----------
        f: Callable
            Function to decorate. Can only accept and return python literals.
        minify: bool
            Minify ``cmd`` code prior to sending.
            Defaults to ``True``.
        register: bool
            Assign an attribute to ``self.task`` with same name as ``f``.
            Defaults to ``True``.
        record: bool
            Each invocation of the executer is recorded for playback upon reconnect.
            Only recommended to be set to ``True`` for a setup-like function.
            Defaults to ``False``.
        implementation: str
            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
            Several methods of the same name can be overloaded that support different implementations.
            Common values include "micropython", and "circuitpython".
            Defaults to an empty string, which **all** implementations will match to.
        """  # noqa: D400
        if f is None:
            return wraps_partial(Device.task, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]

        f.__belay__ = MethodMetadata(executer=TaskExecuter, implementation=implementation, kwargs=kwargs)
        return f

    @overload
    @staticmethod
    def thread(f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    @staticmethod
    def thread(*, implementation: str = "", **kwargs) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    @staticmethod
    def thread(
        f: Optional[Callable[P, R]] = None, implementation: str = "", **kwargs
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        """Spawn on-device thread that executes decorated function.

        Can either be used as a staticmethod ``@Device.thread`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.thread`` for marking functions to a specific ``Device`` instance.

        Parameters
        ----------
        f: Callable
            Function to decorate. Can only accept python literals as arguments.
        minify: bool
            Minify ``cmd`` code prior to sending.
            Defaults to ``True``.
        register: bool
            Assign an attribute to ``self.thread`` with same name as ``f``.
            Defaults to ``True``.
        record: bool
            Each invocation of the executer is recorded for playback upon reconnect.
            Defaults to ``True``.
        implementation: str
            If supplied, the provided method will **only** be used if the board's implementation **name** matches.
            Several methods of the same name can be overloaded that support different implementations.
            Common values include "micropython", and "circuitpython".
            Defaults to an empty string, which **all** implementations will match to.
        """  # noqa: D400
        if f is None:
            return wraps_partial(Device.task, implementation=implementation, **kwargs)  # type: ignore[reportGeneralTypeIssues]
        f.__belay__ = MethodMetadata(executer=ThreadExecuter, implementation=implementation, kwargs=kwargs)
        return f

    def terminal(self, *, exit_char=chr(0x1D)):
        """Start a blocking interactive terminal over the serial port."""
        self._board.exit_raw_repl()  # In case we were previously in raw repl mode.
        miniterm = Miniterm(self._board.serial)
        miniterm.set_rx_encoding("UTF-8")
        miniterm.set_tx_encoding("UTF-8")
        miniterm.exit_character = exit_char
        miniterm.start()
        with contextlib.suppress(KeyboardInterrupt):
            miniterm.join(True)
        miniterm.join()

    def soft_reset(self):
        """Reset device, executing ``main.py`` if available."""
        # When in Raw REPL, ctrl-d will perform a reset, but won't execute ``main.py``
        # https://github.com/micropython/micropython/issues/2249
        self._board.exit_raw_repl()
        self._board.read_until(b">>>")
        self._board.ctrl_d()

    def _traceback_execute(
        self,
        src_file: PathType,
        src_lineno: int,
        name: str,
        cmd: str,
        record: bool = True,
    ):
        """Invoke ``cmd``, and reinterprets raised stacktrace in ``PyboardException``.

        Parameters
        ----------
        src_file: Union[str, Path]
            Path to the file containing the code of the function that ``cmd`` will execute.
        src_lineno: int
            Line number into ``src_file`` that the function starts.
        name: str
            Name of the function.
        cmd: str
            Python command that executes a function on-device.
        record: bool
            Record the call for state-reconstruction if device is accidentally reset.
            Defaults to ``True``.

        Returns
        -------
            Correctly interpreted return value from executing code on-device.
        """
        src_file = str(src_file)

        try:
            res = self(cmd, record=record)
        except PyboardException as e:
            new_lines = []

            msg = e.args[0]
            lines = msg.split("\n")
            for line in lines:
                new_lines.append(line)

                try:
                    file, lineno, fn = line.strip().split(",", 2)
                except ValueError:
                    continue

                if file != 'File "<stdin>"' or fn != f" in {name}":
                    continue

                lineno = int(lineno[6:]) - 1 + src_lineno

                new_lines[-1] = f'  File "{src_file}", line {lineno},{fn}'

                # Get what that line actually is.
                new_lines.append("    " + linecache.getline(src_file, lineno).strip())
            new_msg = "\n".join(new_lines)
            e.args = (new_msg,)
            raise
        return res



"""
The MIT License (MIT).

Copyright (c) 2016 Damien P. George
Copyright (c) 2016 Paul Sokolovsky
Copyright (c) 2022 Jim Mussared

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
import errno
import socket
import struct
import sys
from collections import deque
from pathlib import Path

from .exceptions import AuthenticationError

# Treat this remote directory as a root for file transfers
SANDBOX = ""
# SANDBOX = "/tmp/webrepl/"
DEBUG = 0

WEBREPL_REQ_S = "<2sBBQLH64s"
WEBREPL_PUT_FILE = 1
WEBREPL_GET_FILE = 2
WEBREPL_GET_VER = 3


def debugmsg(msg):
    if DEBUG:
        print(msg)


class WebreplError(Exception):
    pass


class Websocket:
    def __init__(self, s: socket.socket):
        self.s = s
        self.buf = b""

    def write(self, data):
        data_len = len(data)
        hdr = struct.pack(">BB", 130, data_len) if data_len < 126 else struct.pack(">BBH", 130, 126, data_len)
        self.s.send(hdr)
        self.s.send(data)

    def writetext(self, data: bytes):
        data_len = len(data)
        hdr = struct.pack(">BB", 129, data_len) if data_len < 126 else struct.pack(">BBH", 129, 126, data_len)
        self.s.send(hdr)
        self.s.send(data)

    def recvexactly(self, sz):
        res = b""
        while sz:
            data = self.s.recv(sz)
            if not data:
                break
            res += data
            sz -= len(data)
        return res

    def read(self, size, text_ok=False, size_match=True):
        if not self.buf:
            while True:
                hdr = self.recvexactly(2)
                if len(hdr) != 2:
                    raise WebreplError
                fl, sz = struct.unpack(">BB", hdr)
                if sz == 126:
                    hdr = self.recvexactly(2)
                    if len(hdr) != 2:
                        raise WebreplError
                    (sz,) = struct.unpack(">H", hdr)
                if fl == 0x82:
                    break
                if text_ok and fl == 0x81:
                    break
                debugmsg("Got unexpected websocket record of type %x, skipping it" % fl)
                while sz:
                    skip = self.s.recv(sz)
                    debugmsg("Skip data: %s" % skip)
                    sz -= len(skip)
            data = self.recvexactly(sz)
            if len(data) != sz:
                raise WebreplError
            self.buf = data

        d = self.buf[:size]
        self.buf = self.buf[size:]
        if size_match and len(d) != size:
            raise WebreplError
        return d

    def ioctl(self, req, val):
        if req != 9 and val != 2:
            raise WebreplError


def login(ws, passwd):
    while True:
        c = ws.read(1, text_ok=True)
        if c == b":":
            if ws.read(1, text_ok=True) != b" ":
                raise WebreplError
            break
    ws.write(passwd.encode("utf-8") + b"\r")


def read_resp(ws):
    data = ws.read(4)
    sig, code = struct.unpack("<2sH", data)
    if sig != b"WB":
        raise WebreplError
    return code


def send_req(ws, op, sz=0, fname=b""):
    rec = struct.pack(WEBREPL_REQ_S, b"WA", op, 0, 0, sz, len(fname), fname)
    debugmsg("%r %d" % (rec, len(rec)))
    ws.write(rec)


def get_ver(ws):
    send_req(ws, WEBREPL_GET_VER)
    d = ws.read(3)
    d = struct.unpack("<BBB", d)
    return d


def put_file(ws, local_file, remote_file):
    local_file = Path(local_file)
    sz = local_file.stat().st_size
    dest_fname = (SANDBOX + remote_file).encode("utf-8")
    rec = struct.pack(WEBREPL_REQ_S, b"WA", WEBREPL_PUT_FILE, 0, 0, sz, len(dest_fname), dest_fname)
    debugmsg("%r %d" % (rec, len(rec)))
    ws.write(rec[:10])
    ws.write(rec[10:])
    if read_resp(ws) != 0:
        raise WebreplError
    cnt = 0
    with local_file.open("rb") as f:
        while True:
            sys.stdout.write("Sent %d of %d bytes\r" % (cnt, sz))
            sys.stdout.flush()
            buf = f.read(1024)
            if not buf:
                break
            ws.write(buf)
            cnt += len(buf)
    print()
    if read_resp(ws) != 0:
        raise WebreplError


def get_file(ws, local_file, remote_file):
    local_file = Path(local_file)
    src_fname = (SANDBOX + remote_file).encode("utf-8")
    rec = struct.pack(WEBREPL_REQ_S, b"WA", WEBREPL_GET_FILE, 0, 0, 0, len(src_fname), src_fname)
    debugmsg("%r %d" % (rec, len(rec)))
    ws.write(rec)
    if read_resp(ws) != 0:
        raise WebreplError
    with local_file.open("wb") as f:
        cnt = 0
        while True:
            ws.write(b"\0")
            (sz,) = struct.unpack("<H", ws.read(2))
            if sz == 0:
                break
            while sz:
                buf = ws.read(sz)
                if not buf:
                    raise OSError()
                cnt += len(buf)
                f.write(buf)
                sz -= len(buf)
                sys.stdout.write("Received %d bytes\r" % cnt)
                sys.stdout.flush()
    print()
    if read_resp(ws) != 0:
        raise WebreplError


def help(rc=0):
    exename = sys.argv[0].rsplit("/", 1)[-1]
    print("%s - Perform remote file operations using MicroPython WebREPL protocol" % exename)
    print("Arguments:")
    print("  [-p password] <host>:<remote_file> <local_file> - Copy remote file to local file")
    print("  [-p password] <local_file> <host>:<remote_file> - Copy local file to remote file")
    print("Examples:")
    print("  %s script.py 192.168.4.1:/another_name.py" % exename)
    print("  %s script.py 192.168.4.1:/app/" % exename)
    print("  %s -p password 192.168.4.1:/app/script.py ." % exename)
    sys.exit(rc)


def error(msg):
    print(msg)
    sys.exit(1)


def parse_remote(remote):
    host, fname = remote.rsplit(":", 1)
    if fname == "":
        fname = "/"
    port = 8266
    if ":" in host:
        host, port = host.split(":")
        port = int(port)
    return (host, port, fname)


def client_handshake(sock):
    """Simplified client handshake.

    Works for MicroPython's websocket server implementation, but probably not
    for other servers.
    """
    cl = sock.makefile("rwb", 0)
    cl.write(
        b"""\
GET / HTTP/1.1\r
Host: echo.websocket.org\r
Connection: Upgrade\r
Upgrade: websocket\r
Sec-WebSocket-Key: foo\r
\r
"""
    )
    line = cl.readline()
    while 1:
        line = cl.readline()
        if line == b"\r\n":
            break


class WebsocketClosedError(Exception):
    """Attempted to use a closed websocket."""


class WebreplToSerial:
    def __init__(self, uri, password, read_timeout=None):
        self.fifo = deque()
        self.read_timeout = read_timeout

        if uri.startswith("ws://"):
            uri = uri[5:]
        host, *remain = uri.split(":", 1)
        port = int(remain[0]) if remain else 8266

        self.s = socket.socket()
        self.s.settimeout(read_timeout)
        self.s.connect((host, port))
        client_handshake(self.s)

        self.ws = Websocket(self.s)

        login(self.ws, password)
        response = self.read(1024)
        if response == b"\r\nAccess denied\r\n":
            raise AuthenticationError(f"Incorrect password: {repr(password)}")
        elif response != b"\r\nWebREPL connected\r\n>>> ":
            raise WebreplError(f"Unknown login response: {response}")

    def close(self):
        if self.s is not None:
            self.s.close()
        self.s = self.ws = None

    def write(self, data: bytes) -> int:
        if self.ws is None:
            raise WebsocketClosedError
        self.ws.writetext(data)
        return len(data)

    def read(self, size=1) -> bytes:
        if self.ws is None:
            raise WebsocketClosedError

        readin = self.ws.read(size, text_ok=True, size_match=False)
        self.fifo.extend(readin)

        data = b""
        while len(data) < size and len(self.fifo) > 0:
            data += bytes([self.fifo.popleft()])
        return data

    @property
    def in_waiting(self):
        if self.s is None or self.ws is None:
            raise WebsocketClosedError

        n_waiting = len(self.fifo) + len(self.ws.buf)
        if not n_waiting:
            self.s.setblocking(False)
            try:
                n_waiting = len(self.s.recv(1024, socket.MSG_PEEK))
            except BlockingIOError:
                pass
            except OSError as e:
                if e == errno.EAGAIN or e == errno.EWOULDBLOCK:
                    pass
                else:
                    raise
            self.s.setblocking(True)
            return n_waiting
        else:
            return n_waiting



import importlib.resources
import secrets
import string
from functools import lru_cache, partial, wraps

from . import snippets

_python_identifier_chars = string.ascii_uppercase + string.ascii_lowercase + string.digits


def wraps_partial(f, *args, **kwargs):
    """Wrap and partial of a function."""
    return wraps(f)(partial(f, *args, **kwargs))


def random_python_identifier(n=16):
    return "_" + "".join(secrets.choice(_python_identifier_chars) for _ in range(n))


@lru_cache
def read_snippet(name):
    resource = f"{name}.py"
    return importlib.resources.files(snippets).joinpath(resource).read_text()



import inspect
from abc import abstractmethod
from functools import wraps
from typing import Callable, Optional, TypeVar, Union, overload

from autoregistry import Registry
from typing_extensions import ParamSpec

from ._minify import minify as _minify
from .exceptions import FeatureUnavailableError, SpecialFunctionNameError
from .helpers import random_python_identifier, wraps_partial
from .inspect import getsource
from .typing import BelayCallable

P = ParamSpec("P")
R = TypeVar("R")


class Executer(Registry, suffix="Executer"):
    def __init__(self, device):
        # Use object.__setattr__ to avoid Executer.__setattr__ raising an error
        object.__setattr__(self, "_belay_device", device)
        object.__setattr__(self, "_belay_executers", [])

    def __setattr__(self, name: str, value: BelayCallable):
        if name.startswith("_belay") or name.startswith("__belay") or (name.startswith("__") and name.endswith("__")):
            raise SpecialFunctionNameError(f'Not allowed to register function named "{name}".')
        super().__setattr__(name, value)

    def __getattr__(self, name: str) -> BelayCallable:
        # Just here for linting purposes.
        raise AttributeError

    @abstractmethod
    def __call__(self):
        raise NotImplementedError


class _GlobalExecuter(Executer, skip=True):
    @overload
    def __call__(self, f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    def __call__(
        self,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = True,
        ignore_errors: bool = False,
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    def __call__(
        self,
        f: Optional[Callable[P, R]] = None,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = True,
        ignore_errors: bool = False,
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        if f is None:
            return wraps_partial(self, minify=minify, register=register, record=record)
        if inspect.isgeneratorfunction(f):
            raise ValueError(f"@Device.{type(self).__registry__.name} does not support generators.")
        name = f.__name__
        src_code, src_lineno, src_file = getsource(f, strip_signature=True)
        if minify:
            src_code = _minify(src_code)
        signature = inspect.signature(f)

        @wraps(f)
        def executer(*args, **kwargs):
            cmd = src_code
            bound_arguments = signature.bind(*args, **kwargs)
            bound_arguments.apply_defaults()
            arg_assign_cmd = "\n".join(f"{name}={repr(val)}" for name, val in bound_arguments.arguments.items())
            if arg_assign_cmd:
                cmd = arg_assign_cmd + "\n" + cmd

            try:
                self._belay_device._traceback_execute(src_file, src_lineno, name, cmd, record=record)
            except Exception:
                if not ignore_errors:
                    raise

        if register:
            setattr(self, name, executer)
        self._belay_executers.append(executer)

        return executer


class SetupExecuter(_GlobalExecuter):
    pass


class TeardownExecuter(_GlobalExecuter):
    pass


class TaskExecuter(Executer):
    @overload
    def __call__(self, f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    def __call__(
        self,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = False,
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    def __call__(
        self,
        f: Optional[Callable[P, R]] = None,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = False,
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        """See ``Device.task``."""
        if f is None:
            return wraps_partial(self, minify=minify, register=register, record=record)

        name = f.__name__
        src_code, src_lineno, src_file = getsource(f)

        # Send the source code over to the device.
        self._belay_device(src_code, minify=minify)

        @wraps(f)
        def func_executer(*args, **kwargs):
            cmd = f"{name}(*{repr(args)}, **{repr(kwargs)})"

            return self._belay_device._traceback_execute(src_file, src_lineno, name, cmd, record=record)

        @wraps(f)
        def gen_executer(*args, **kwargs):
            if record:
                raise NotImplementedError("Recording of generator tasks is currently not supported.")
            # Step 1: Create the on-device generator
            gen_identifier = random_python_identifier()
            cmd = f"{gen_identifier} = {name}(*{repr(args)}, **{repr(kwargs)})"
            self._belay_device._traceback_execute(src_file, src_lineno, name, cmd, record=False)
            # Step 2: Create the host generator that invokes ``next()`` on-device.

            def gen_inner():
                send_val = None
                try:
                    while True:
                        cmd = f"__belay_next({gen_identifier}, {repr(send_val)})"
                        send_val = yield self._belay_device._traceback_execute(
                            src_file, src_lineno, name, cmd, record=False
                        )
                except StopIteration:
                    pass
                # Delete the exhausted generator on-device.
                self._belay_device(f"del {gen_identifier}")

            return gen_inner()

        executer = gen_executer if inspect.isgeneratorfunction(f) else func_executer

        if register:
            setattr(self, name, executer)
        self._belay_executers.append(executer)

        return executer


class ThreadExecuter(Executer):
    @overload
    def __call__(self, f: Callable[P, R]) -> Callable[P, R]:
        ...

    @overload
    def __call__(
        self,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = True,
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        ...

    def __call__(
        self,
        f: Optional[Callable[P, R]] = None,
        *,
        minify: bool = True,
        register: bool = True,
        record: bool = True,
    ) -> Union[Callable[[Callable[P, R]], Callable[P, R]], Callable[P, R]]:
        """See ``Device.thread``."""
        if f is None:
            return wraps_partial(self, minify=minify, register=register, record=record)

        if self._belay_device.implementation.name == "circuitpython":
            raise FeatureUnavailableError("CircuitPython does not support threading.")

        name = f.__name__
        src_code, src_lineno, src_file = getsource(f)

        # Send the source code over to the device.
        self._belay_device(src_code, minify=minify)

        @wraps(f)
        def executer(*args, **kwargs):
            cmd = f"import _thread; _thread.start_new_thread({name}, {repr(args)}, {repr(kwargs)})"
            self._belay_device._traceback_execute(src_file, src_lineno, name, cmd, record=record)

        if register:
            setattr(self, name, executer)
        self._belay_executers.append(executer)

        return executer



import os, time, board, digitalio
from time import sleep
try:
    import analogio
except ImportError:
    pass
try:
    from busio import I2C
except ImportError:
    pass
try:
    from busio import SPI
except ImportError:
    pass



"""Pydantic models for validation Belay configuration.
"""

from functools import partial
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import BaseModel as PydanticBaseModel
from pydantic import validator

validator_reuse = partial(validator, allow_reuse=True)
prevalidator_reuse = partial(validator_reuse, pre=True)


class BaseModel(PydanticBaseModel):
    class Config:
        allow_mutation = False


class DependencySourceConfig(BaseModel):
    uri: str
    develop: bool = False  # If true, local dependency is in "editable" mode.

    rename_to_init: bool = False


DependencyList = List[DependencySourceConfig]


def _dependencies_name_validator(dependencies) -> dict:
    for group_name in dependencies:
        if not group_name.isidentifier():
            raise ValueError("Dependency group name must be a valid python identifier.")
    return dependencies


def _dependencies_preprocessor(dependencies) -> Dict[str, List[dict]]:
    """Preprocess various dependencies based on dtype.

    * ``str`` -> single dependency that may get renamed to __init__.py, if appropriate.
    * ``list`` -> list of dependencies. If an element is a str, it will not
      get renamed to __init__.py.
    * ``dict`` -> full dependency specification.
    """
    out = {}
    for group_name, group_value in dependencies.items():
        if isinstance(group_value, str):
            group_value = [
                {
                    "uri": group_value,
                    "rename_to_init": True,
                }
            ]
        elif isinstance(group_value, list):
            group_value_out = []
            for elem in group_value:
                if isinstance(elem, str):
                    group_value_out.append(
                        {
                            "uri": elem,
                        }
                    )
                elif isinstance(elem, list):
                    raise TypeError("Cannot have double nested lists in dependency specification.")
                elif isinstance(elem, (dict, DependencySourceConfig)):
                    group_value_out.append(elem)
                else:
                    raise NotImplementedError
            group_value = group_value_out
        elif isinstance(group_value, dict):
            group_value = group_value.copy()
            group_value.setdefault("rename_to_init", True)
            group_value = [group_value]
        elif isinstance(group_value, DependencySourceConfig):
            # Nothing to do
            pass
        else:
            raise TypeError

        out[group_name] = group_value

    return out


def walk_dependencies(packages: dict):
    """Walks over all package/dependency pairs.

    Yields
    ------
    tuple
        (package_name, dependency)
    """
    for package_name, dependencies in packages.items():
        for dependency in dependencies:
            yield package_name, dependency


class GroupConfig(BaseModel):
    optional: bool = False
    dependencies: Dict[str, DependencyList] = {}

    ##############
    # VALIDATORS #
    ##############
    _v_dependencies_preprocessor = prevalidator_reuse("dependencies")(_dependencies_preprocessor)
    _v_dependencies_names = validator_reuse("dependencies")(_dependencies_name_validator)

    @validator("dependencies")
    def max_1_rename_to_init(packages: dict):
        rename_to_init_count = {}
        for package_name, dependency in walk_dependencies(packages):
            rename_to_init_count.setdefault(package_name, 0)
            rename_to_init_count[package_name] += dependency.rename_to_init
            if rename_to_init_count[package_name] > 1:
                raise ValueError(f'{package_name} has more than 1 dependency marked with "rename_to_init".')
        return packages


class BelayConfig(BaseModel):
    """Configuration schema under the ``tool.belay`` section of ``pyproject.toml``."""

    # Name/Folder of project's primary micropython code.
    name: Optional[str] = None

    # Items in project directory to ignore.
    ignore: Optional[list] = []

    # "main" dependencies
    dependencies: Dict[str, DependencyList] = {}

    # Path to where dependency groups should be stored relative to project's root.
    dependencies_path: Path = Path(".belay/dependencies")

    # Other dependencies
    group: Dict[str, GroupConfig] = {}

    ##############
    # VALIDATORS #
    ##############
    _v_dependencies_preprocessor = prevalidator_reuse("dependencies")(_dependencies_preprocessor)
    _v_dependencies_names = validator_reuse("dependencies")(_dependencies_name_validator)

    @validator("group")
    def main_not_in_group(cls, v):
        if "main" in v:
            raise ValueError(
                'Specify "main" group dependencies under "tool.belay.dependencies", '
                'not "tool.belay.group.main.dependencies"'
            )
        return v



import asyncio
import contextlib

import questionary
import typer
from questionary import Choice

from belay import Device, DeviceMeta
from belay.cli.questionary_ext import press_any_key_to_continue, select_table
from belay.usb_specifier import list_devices


async def blink_loop(device):
    while True:
        device.led(True)
        await asyncio.sleep(0.5)
        device.led(False)
        await asyncio.sleep(0.5)


async def blink_until_prompt(device):
    blink_task = asyncio.create_task(blink_loop(device))
    await press_any_key_to_continue().ask_async()
    blink_task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await blink_task


class CircuitPythonBlinker(metaclass=DeviceMeta):
    @Device.setup(implementation="circuitpython")
    def setup(pin: str, is_neopixel) -> None:
        import board
        import digitalio

        if is_neopixel:
            from neopixel_write import neopixel_write

        pin_name = "GP"
        pin_name = "LED" if pin.upper() == "LED" else f"GP{pin}"

        led_io = digitalio.DigitalInOut(getattr(board, pin_name))
        led_io.direction = digitalio.Direction.OUTPUT

    @Device.task(implementation="circuitpython")
    def led(value: bool) -> None:
        if is_neopixel:
            val = (255, 255, 255) if value else (0, 0, 0)
            neopixel_write(led_io, bytearray(val))
        else:
            led_io.value = value

    @Device.teardown(implementation="circuitpython", ignore_errors=True)
    def teardown():
        if is_neopixel:
            neopixel_write(led_io, b"\x00\x00\x00")
        else:
            led_io.value = False


class MicroPythonBlinker(metaclass=DeviceMeta):
    @Device.setup
    def setup(pin, is_neopixel) -> None:
        from machine import Pin

        led_io = Pin(int(pin), Pin.OUT)

        if is_neopixel:
            from neopixel import NeoPixel

            pixel = NeoPixel(led_io, 1)

    @Device.task
    def led(value: bool) -> None:
        if is_neopixel:
            pixel[0] = (255, 255, 255) if value else (0, 0, 0)
            pixel.write()
        else:
            led_io(value)

    @Device.teardown(ignore_errors=True)
    def teardown():
        if is_neopixel:
            pixel[0] = (0, 0, 0)
            pixel.write()
        else:
            led_io(False)


class Blinker(Device, CircuitPythonBlinker, MicroPythonBlinker):
    """MicroPython + CircuitPython led/neopixel interactions."""


def select():
    """Interactive board selector.

    For determining board-specific metadata for repeatable connections.
    """
    style = "bold"
    table_spec = {
        "vid": 6,
        "pid": 6,
        "serial_number": 18,
        "manufacturer": 18,
        "product": 18,
        "location": 18,
    }
    header = " ".join(f"{k:{v}}" for k, v in table_spec.items())
    specs, choices = [], []
    for spec in list_devices():
        specs.append(spec)
        choices.append(
            Choice(
                " ".join(f"{str(getattr(spec, k)):{v}.{v}}" for k, v in table_spec.items()),
                value=len(choices),
            )
        )

    if not choices:
        print("Detected no devices.")
        raise typer.Exit(code=1)

    device_index = select_table(
        "Select USB Device (Use arrow keys):",
        header,
        choices,
    ).ask()

    spec = specs[device_index]

    with Blinker(spec) as device:
        questionary.print(str(device.implementation), style=style)

        def validate_led_pin_number(response):
            if not response:
                return True

            try:
                int(response)
            except ValueError:
                return False
            return True

        pin_number = questionary.text(
            "Blink LED Pin Number [skip]?",
            validate=validate_led_pin_number,
        ).ask()

        if pin_number:
            is_neopixel = questionary.confirm("Is this a NeoPixel?", default=False).ask()
            device.setup(pin_number, is_neopixel)

            asyncio.run(blink_until_prompt(device))

    spec_json = spec.json(exclude_none=True)
    questionary.print("\n")
    questionary.print("Either set the BELAY_DEVICE environment variable:", style=style)
    questionary.print(f"    export BELAY_DEVICE='{spec_json}'")
    questionary.print("And in python code, instantiate Device without arguments:", style=style)
    questionary.print("    device = belay.Device()")
    questionary.print("")
    questionary.print("Or, add the following (or a subset) to your python code:", style=style)
    questionary.print(f"    spec = belay.{spec!r}\n    device = belay.Device(spec)\n")



import builtins
import contextlib

with contextlib.suppress(ImportError):
    import readline
import shutil

import typer
from typer import Argument, Option, Typer

from belay.project import find_cache_folder

app = Typer(no_args_is_help=True, help="Perform action's on Belay's cache.")


@app.command()
def clear(
    prefix: str = Argument("", help="Clear all caches that start with this."),
    yes: bool = Option(
        False,
        "--yes",
        "-y",
        help='Automatically answer "yes" to all confirmation prompts.',
    ),
    all: bool = Option(False, "--all", "-a", help="Clear all caches."),
):
    """Clear cache."""
    if (not prefix and not all) or (prefix and all):
        print('Either provide a prefix OR set the "--all" flag.')
        raise typer.Exit()

    cache_folder = find_cache_folder()

    prefix += "*"
    cache_paths = builtins.list(cache_folder.glob(prefix))
    cache_names = [x.name for x in cache_paths]

    if not cache_paths:
        print(f'No caches found starting with "{prefix}"')
        raise typer.Exit()

    if not yes:
        print("Found caches:")
        for cache_name in cache_names:
            print(f"  • {cache_name}")
        typer.confirm("Clear these caches?", abort=True)

    for path in cache_paths:
        if path.is_file():
            path.unlink()
        else:
            shutil.rmtree(path)


@app.command()
def list():
    """List cache elements."""
    cache_folder = find_cache_folder()
    items = [x.name for x in cache_folder.glob("*")]

    for item in items:
        print(item)


@app.command()
def info():
    """Display cache location and size."""
    cache_folder = find_cache_folder()

    print(f"Location: {cache_folder}")

    n_elements = len(builtins.list(cache_folder.glob("*")))
    print(f"Elements: {n_elements}")

    size_in_bytes = sum(f.stat().st_size for f in cache_folder.glob("**/*") if f.is_file())
    size_in_megabytes = size_in_bytes / (1 << 20)
    print(f"Total Size: {size_in_megabytes:0.3}MB")



import platform
from functools import lru_cache
from pathlib import Path
from typing import List, Union

import tomli

from belay.packagemanager import BelayConfig, Group


@lru_cache
def find_pyproject() -> Path:
    path = Path("pyproject.toml").absolute()

    for parent in path.parents:
        candidate = parent / path.name
        if candidate.exists():
            return candidate
    raise FileNotFoundError(
        f'Cannot find a pyproject.toml in the current directory "{Path().absolute()}" or any parent directory.'
    )


@lru_cache
def find_project_folder() -> Path:
    return find_pyproject().parent


@lru_cache
def find_belay_folder() -> Path:
    return find_project_folder() / ".belay"


@lru_cache
def find_dependencies_folder() -> Path:
    config = load_pyproject()
    return find_project_folder() / config.dependencies_path


@lru_cache
def find_cache_folder() -> Path:
    system = platform.system()
    cache_folder = Path.home()

    if system == "Windows":
        cache_folder /= "AppData/Local/belay/Cache"
    elif system == "Darwin":
        cache_folder /= "Library/Caches/belay"
    else:
        cache_folder /= ".cache/belay"

    return cache_folder.absolute()


@lru_cache
def find_cache_dependencies_folder() -> Path:
    return find_cache_folder() / "dependencies"


@lru_cache
def load_toml(path: Union[str, Path]) -> dict:
    path = Path(path)
    with path.open("rb") as f:
        toml = tomli.load(f)

    try:
        toml = toml["tool"]["belay"]
    except KeyError:
        return {}

    return toml


@lru_cache
def load_pyproject() -> BelayConfig:
    """Load the pyproject TOML file."""
    pyproject_path = find_pyproject()
    belay_data = load_toml(pyproject_path)
    return BelayConfig(**belay_data)


@lru_cache
def load_groups() -> List[Group]:
    config = load_pyproject()
    groups = [Group("main", dependencies=config.dependencies)]
    groups.extend(Group(name, **definition.dict()) for name, definition in config.group.items())
    groups.sort(key=lambda x: x.name)
    return groups
